# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import warnings
import torch
import logging
from terratorch.registry import TERRATORCH_FULL_MODEL_REGISTRY
from huggingface_hub import hf_hub_download

logger = logging.getLogger("terramind")

try:
    from .vqvae import VQVAE, DiVAE
    vqvae_available = True
    import_error = None
except Exception as e:
    logger.debug(f"Could not import TerraMind tokenizer due to ImportError({e})")
    vqvae_available = False
    import_error = e

try:
    from .text.text_tokenizer import CoordsTokenizer, CaptionTokenizer
    tokenizers_available = True
    import_error_tokenizers = None
except Exception as e:
    logger.debug(f"Could not import tokenizers due to ImportError({e})")
    tokenizers_available = False
    import_error_tokenizers = e


# Model definitions
__all__ = [
    "terramind_v1_tokenizer_s2l2a",
    "terramind_v1_tokenizer_s1rtc",
    "terramind_v1_tokenizer_s1grd",
    "terramind_v1_tokenizer_dem",
    "terramind_v1_tokenizer_lulc",
    "terramind_v1_tokenizer_ndvi",
    "terramind_v01_tokenizer_s2l2a",
    "terramind_v01_tokenizer_s1grd",
    "terramind_v01_tokenizer_dem",
    "terramind_v01_tokenizer_lulc",
    "terramind_v01_caption_tokenizer",
    "terramind_v1_coords_tokenizer",
]

pretrained_weights = {
    "terramind_v1_tokenizer_s2l2a": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-S2L2A",
        "hf_hub_filename": "TerraMind_Tokenizer_S2L2A.pt",
    },
    "terramind_v1_tokenizer_s1rtc": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-S1RTC",
        "hf_hub_filename": "TerraMind_Tokenizer_S1RTC.pt",
    },
    "terramind_v1_tokenizer_s1grd": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-S1GRD",
        "hf_hub_filename": "TerraMind_Tokenizer_S1GRD.pt",
    },
    "terramind_v1_tokenizer_dem": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-DEM",
        "hf_hub_filename": "TerraMind_Tokenizer_DEM.pt",
    },
    "terramind_v1_tokenizer_lulc": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-LULC",
        "hf_hub_filename": "TerraMind_Tokenizer_LULC.pt",
    },
    "terramind_v1_tokenizer_ndvi": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-NDVI",
        "hf_hub_filename": "TerraMind_Tokenizer_NDVI.pt",
    },
    "terramind_v1_coords_tokenizer": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-Coords",
        "hf_hub_filename": "config.json",
    },
    "terramind_v01_tokenizer_s2l2a": {
        "hf_hub_id": "FAST-EO/TerraMind-v0.1-Tokenizer-S2L2A",
        "hf_hub_filename": "TerraMind_v01_Tokenizer_S2L2A.pt",
    },
    "terramind_v01_tokenizer_s1grd": {
        "hf_hub_id": "FAST-EO/TerraMind-v0.1-Tokenizer-S1GRD",
        "hf_hub_filename": "TerraMind_v01_Tokenizer_S1GRD.pt",
    },
    "terramind_v01_tokenizer_dem": {
        "hf_hub_id": "FAST-EO/TerraMind-v0.1-Tokenizer-DEM",
        "hf_hub_filename": "TerraMind_v01_Tokenizer_DEM.pt",
    },
    "terramind_v01_tokenizer_lulc": {
        "hf_hub_id": "FAST-EO/TerraMind-v0.1-Tokenizer-LULC",
        "hf_hub_filename": "TerraMind_v01_Tokenizer_LULC.pt",
    },
    "terramind_v01_caption_tokenizer": {
        "hf_hub_id": "FAST-EO/TerraMind-v0.1-Tokenizer-Text",
        "hf_hub_filename": "config.json",
    },
}

def build_vqvae(
        model_type: str = "divae",
        variant: str = None,
        pretrained: bool = False,
        ckpt_path: str | None = None,
        **kwargs):

    if not vqvae_available:
        warnings.warn(f"Cannot import VQVAE/DiVAE from terramind. "
                      f"\nMake sure to install `pip install diffusers==0.30.0`.")
        raise import_error

    if model_type == "divae":
        model = DiVAE(**kwargs)
    elif model_type == "vqvae":
        model = VQVAE(**kwargs)
    else:
        raise ValueError(f"Unknown model type: {model_type}")

    if ckpt_path is not None:
        # Load model from checkpoint
        state_dict = torch.load(ckpt_path, map_location="cpu", weights_only=True)
        loaded_keys = model.load_state_dict(state_dict, strict=False)
        if loaded_keys.missing_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")
        if loaded_keys.unexpected_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")

    elif pretrained:
        # Load model from Hugging Face
        state_dict_file = hf_hub_download(repo_id=pretrained_weights[variant]["hf_hub_id"],
                                          filename=pretrained_weights[variant]["hf_hub_filename"])
        state_dict = torch.load(state_dict_file, map_location="cpu", weights_only=True)
        model.load_state_dict(state_dict, strict=True)

    return model


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_s2l2a(**kwargs):
    """
    S2L2A Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=224, input_size_max=256, resolution_step=32, input_size_enc=None, encoder_type="vit_b_enc", decoder_type="unet_patched", post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, from_scratch=True, dec_transformer_dropout=0.2, quantizer_type="fsq", codebook_size="8-8-8-6-5", latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy="batch_random", commitment_weight=1.0, kmeans_init=False, num_train_timesteps=1000, prediction_type="sample", beta_schedule="linear", zero_terminal_snr=True, cls_free_guidance_dropout=0.0, masked_cfg=False, masked_cfg_low=0, masked_cfg_high=None, thresholding=True, loss_fn="mse", conditioning="concat", resolution_cond=False, eval_res_cond=None, batch_size=1, accum_grad=1, batch_size_eval=None, epochs=100, stop_after_epoch=250, save_ckpt_freq=1, opt="adamw", opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.0064, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype="fp16", model_ema=True, model_ema_decay=0.9999, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain="sen2l2a@264", mask_value=None, data_path=["./data/MajorTOM-Core/train", "./data/SSL4EOS12/train_single"], eval_data_path="./data/MajorTOM-Core/val", imagenet_default_mean_and_std=False, standardize_surface_normals=False, min_crop_scale=0.8, cache_datasets=False, dataset_size=None, sample_sentinel_data=True, num_seasons=1, num_locations=64, dist_eval=True, step_eval=True, eval_noise_schedule="DDIMScheduler", num_eval_timesteps=50, input_size_eval=[256], num_eval_metrics_samples=50000, eval_freq=2000, eval_metrics_freq=0, eval_image_log_freq=2000, num_logged_images=15, eval_only=False, no_inception=False, log_codebook_usage=True, output_dir="output/tokenization/divae/rgb/Sentinel2-ViTB-UNetP4_16k_224-264/2k_continue", device="cuda", seed=0, resume="output/tokenization/divae/rgb/Sentinel2-ViTB-UNetP4_16k_224-264/2k_continue/checkpoint_29.pth", auto_resume=True, start_epoch=30, num_workers=5, pin_mem=True, find_unused_params=True, log_wandb=True, wandb_project=None, wandb_entity="erik-scheurer", wandb_run_name="2048_S2TerraMesh_divae/rgb/Sentinel2-ViTB-UNetP4_16k_224-264", wandb_tags=[], show_user_warnings=False, world_size=32, local_rank=-1, dist_on_itp=False, dist_url="env://", run_name="tokenization/divae/rgb/Sentinel2-ViTB-UNetP4_16k_224-264", patch_size_dec=4, clip_sample=True, epoch_eval=False, config_path="/p/project/geofm4eo/scheurer2/4m4eo/cfgs/default/tokenization/divae/rgb/Sentinel2-ViTB-UNetP4_16k_224-264.yaml", rank=0, gpu=0, distributed=True, dist_backend="nccl", num_tasks=32, all_domains=["sen2l2a@264"], input_size=256, lr=0.0001414213562373095, effective_batch_size=2048)
    """
    tokenizer = build_vqvae(
        variant="terramind_v1_tokenizer_s2l2a",
        image_size=256,
        n_channels=12,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_s1rtc(**kwargs):
    """
    S1RTC Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=224, input_size_max=256, resolution_step=32, input_size_enc=None, encoder_type="vit_b_enc", decoder_type="unet_patched", post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, from_scratch=True, dec_transformer_dropout=0.2, quantizer_type="fsq", codebook_size="8-8-8-6-5", latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy="batch_random", commitment_weight=1.0, kmeans_init=False, num_train_timesteps=1000, prediction_type="sample", beta_schedule="linear", zero_terminal_snr=True, cls_free_guidance_dropout=0.0, masked_cfg=False, masked_cfg_low=0, masked_cfg_high=None, thresholding=True, loss_fn="mse", conditioning="concat", resolution_cond=False, eval_res_cond=None, batch_size=1, accum_grad=1, batch_size_eval=None, epochs=100, stop_after_epoch=250, save_ckpt_freq=1, opt="adamw", opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.0064, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype="fp16", model_ema=True, model_ema_decay=0.9999, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain="sen1@264", mask_value=None, data_path="./data/MajorTOM-Core/train/", eval_data_path="./data/MajorTOM-Core/val", imagenet_default_mean_and_std=False, standardize_surface_normals=False, min_crop_scale=0.8, cache_datasets=False, dataset_size=None, sample_sentinel_data=True, num_seasons=1, num_locations=64, dist_eval=True, step_eval=True, eval_noise_schedule="DDIMScheduler", num_eval_timesteps=50, input_size_eval=[256], num_eval_metrics_samples=50000, eval_freq=1000, eval_metrics_freq=0, eval_image_log_freq=1000, num_logged_images=5, eval_only=False, no_inception=False, log_codebook_usage=True, output_dir="output/tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264/2k_MajorTOM224256", device="cuda", seed=0, resume="output/tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264/2k_MajorTOM224256/checkpoint_14.pth", auto_resume=True, start_epoch=15, num_workers=5, pin_mem=True, find_unused_params=True, log_wandb=True, wandb_project=None, wandb_entity="erik-scheurer", wandb_run_name="2048_2k_MajorTOM224-256", wandb_tags=[], show_user_warnings=False, world_size=32, local_rank=-1, dist_on_itp=False, dist_url="env://", run_name="tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264", patch_size_dec=4, clip_sample=True, epoch_eval=False, config_path="/p/project/geofm4eo/scheurer2/4m4eo/cfgs/default/tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264.yaml", rank=0, gpu=0, distributed=True, dist_backend="nccl", num_tasks=32, all_domains=["sen1@264"], input_size=256, lr=0.0001414213562373095, effective_batch_size=2048)
    """
    
    tokenizer = build_vqvae(
        variant="terramind_v1_tokenizer_s1rtc",
        image_size=256,
        n_channels=2,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer

@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_s1grd(**kwargs):
    """
    S1GRD Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=224, input_size_max=256, resolution_step=32, input_size_enc=None, encoder_type="vit_b_enc", decoder_type="unet_patched", post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, from_scratch=True, dec_transformer_dropout=0.2, quantizer_type="fsq", codebook_size="8-8-8-6-5", latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy="batch_random", commitment_weight=1.0, kmeans_init=False, num_train_timesteps=1000, prediction_type="sample", beta_schedule="linear", zero_terminal_snr=True, cls_free_guidance_dropout=0.0, masked_cfg=False, masked_cfg_low=0, masked_cfg_high=None, thresholding=True, loss_fn="mse", conditioning="concat", resolution_cond=False, eval_res_cond=None, batch_size=1, batch_size_eval=None, epochs=100, save_ckpt_freq=1, opt="adamw", opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.004266666667, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype="fp16", model_ema=True, model_ema_decay=0.9999, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain="sen1@264", mask_value=None, data_path="../fm-geospatial/data/SSL4EOS12/train/", eval_data_path="../fm-geospatial/data/SSL4EOS12/train/", imagenet_default_mean_and_std=False, standardize_surface_normals=False, min_crop_scale=0.8, cache_datasets=False, use_wds=False, s3_endpoint="", s3_data_endpoint=None, wds_n_repeats=1, wds_shuffle_buffer_tar=1000, wds_shuffle_buffer_repeat=1000, s3_multipart_chunksize_mb=512, s3_multipart_threshold_mb=512, dataset_size=None, sample_sentinel_data=True, num_seasons=1, num_locations=64, dist_eval=True, step_eval=True, eval_noise_schedule="DDIMScheduler", num_eval_timesteps=50, input_size_eval=[256], num_eval_metrics_samples=5, eval_freq=5000, eval_metrics_freq=5000, eval_image_log_freq=5000, num_logged_images=15, eval_only=False, no_inception=False, log_codebook_usage=True, output_dir="../fm-geospatial/4m4eo-checkpoints/S1-SSL4EO-full-run/", device="cuda", seed=0, resume="", auto_resume=False, start_epoch=0, num_workers=10, pin_mem=True, find_unused_params=True, log_wandb=True, wandb_project="4m4eo", wandb_entity=None, wandb_run_name="divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264", wandb_tags=[], show_user_warnings=False, world_size=6, local_rank=-1, dist_on_itp=False, dist_url="env://", s3_path="", s3_save_dir="", run_name="tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264", patch_size_dec=4, clip_sample=True, epoch_eval=False, config_path="cfgs/default/tokenization/divae/rgb/Sentinel1-ViTB-UNetP4_16k_224-264.yaml", rank=0, gpu=0, distributed=True, dist_backend="nccl", num_tasks=6, all_domains=["sen1@264"], input_size=256, lr=0.00010000000000781251)
    """
    
    tokenizer = build_vqvae(
        variant="terramind_v1_tokenizer_s1grd",
        image_size=256,
        n_channels=2,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_dem(**kwargs):
    """
    DEM Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=256, input_size_max=256, resolution_step=32, input_size_enc=None, encoder_type="vit_b_enc", decoder_type="unet_patched", post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, from_scratch=True, dec_transformer_dropout=0.2, quantizer_type="fsq", codebook_size="8-8-8-6-5", latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy="batch_random", commitment_weight=1.0, kmeans_init=False, num_train_timesteps=1000, prediction_type="sample", beta_schedule="linear", zero_terminal_snr=True, cls_free_guidance_dropout=0.0, masked_cfg=False, masked_cfg_low=0, masked_cfg_high=None, thresholding=True, loss_fn="mse", conditioning="concat", resolution_cond=False, eval_res_cond=None, batch_size=1, accum_grad=2, batch_size_eval=None, epochs=20, stop_after_epoch=250, save_ckpt_freq=1, opt="adamw", opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.0064, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype="fp16", model_ema=True, model_ema_decay=0.9999, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain="dem@264", mask_value=None, data_path="./data/TerraMesh/train", eval_data_path="./data/TerraMesh/val", imagenet_default_mean_and_std=False, standardize_surface_normals=False, min_crop_scale=0.8, cache_datasets=False, dataset_size=None, sample_sentinel_data=False, num_seasons=1, num_locations=64, dist_eval=True, step_eval=True, eval_noise_schedule="DDIMScheduler", num_eval_timesteps=50, input_size_eval=[256], num_eval_metrics_samples=5, eval_freq=2000, eval_metrics_freq=0, eval_image_log_freq=2000, num_logged_images=5, eval_only=False, no_inception=False, log_codebook_usage=True, output_dir="artifacts/runs/sm-dem-20-epochs", device="cuda", seed=0, resume="artifacts/runs/sm-dem-20-epochs/checkpoint_12.pth", auto_resume=True, start_epoch=13, num_workers=8, pin_mem=True, find_unused_params=True, log_wandb=True, wandb_project="terramind-dem", wandb_entity="fasteo-ibm-jsc", wandb_run_name="2048_epochs_20/divae/rgb/DEM-ViTB-UNetP4_16k_224-264", wandb_run_id=None, wandb_tags=[], show_user_warnings=False, world_size=16, local_rank=-1, dist_on_itp=False, dist_url="env://", run_name="tokenization/divae/rgb/DEM-ViTB-UNetP4_16k_224-264", patch_size_dec=4, clip_sample=True, epoch_eval=False, config_path="/p/home/jusers/maurogiovanni1/juwels/4m4eo/cfgs/default/tokenization/divae/rgb/DEM-ViTB-UNetP4_16k_224-264.yaml", rank=0, gpu=0, distributed=True, dist_backend="nccl", num_tasks=16, all_domains=["dem@264"], input_size=256, lr=0.0001414213562373095, effective_batch_size=2048)
    """
    
    tokenizer = build_vqvae(
        variant="terramind_v1_tokenizer_dem",
        image_size=256,
        n_channels=1,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_lulc(**kwargs):
    """
    LULC Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=224, input_size_max=256, resolution_step=16, input_size_enc=None, input_size_dec=None, encoder_type="vit_b_enc", decoder_type="vit_b_dec", post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, loss_fn="cross_entropy", percept_loss_type=None, percept_loss_blocks="blocks.2-blocks.5-blocks.8-blocks.11", percept_loss_distance="cosine", percept_loss_weight=0.0, quantizer_type="fsq", codebook_size="7-5-5-5-5", num_codebooks=1, latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy="batch_random", commitment_weight=1.0, kmeans_init=False, batch_size=1, accum_grad=2, batch_size_eval=None, epochs=20, save_ckpt_freq=1, opt="adamw", opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.0016, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype="bf16", dtype_percept=None, model_ema=True, model_ema_decay=0.99, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain="lulc@264", mask_value=None, data_path="./data/TerraMesh/train", eval_data_path="./data/TerraMesh/val", min_crop_scale=0.8, cache_datasets=False, num_locations=64, dist_eval=True, step_eval=True, input_size_eval=[256], num_eval_metrics_samples=50000, eval_freq=2000, eval_metrics_freq=0, eval_image_log_freq=1000, num_logged_images=5, eval_only=False, no_inception=False, log_codebook_usage=True, keep_batches_ratio=1.0, output_dir="/p/home/jusers/maurogiovanni1/juwels/4m4eo/artifacts/runs/sm-lulc-20-00016", device="cuda", seed=0, resume="/p/home/jusers/maurogiovanni1/juwels/4m4eo/artifacts/runs/sm-lulc-20-00016/checkpoint_13.pth", auto_resume=True, start_epoch=14, num_workers=8, pin_mem=True, find_unused_params=False, log_wandb=True, wandb_project="terramind-lulc", wandb_entity="fasteo-ibm-jsc", wandb_run_name="2048_epochs_20/lr_00016/acc/vqvae/lulc/LULC_ViTB-ViTB_4k_224-256", wandb_tags=[], show_user_warnings=False, world_size=16, local_rank=-1, dist_on_itp=False, dist_url="env://", run_name="tokenization/vqvae/lulc/LULC_ViTB-ViTB_4k_224-256", stop_after_epoch=250, epoch_eval=False, imagenet_default_mean_and_std=False, sample_sentinel_data=True, num_seasons=1, config_path="/p/home/jusers/maurogiovanni1/juwels/4m4eo/cfgs/default/tokenization/vqvae/lulc/LULC_ViTB-ViTB_4k_224-256.yaml", rank=0, gpu=0, distributed=True, dist_backend="nccl", num_tasks=16, all_domains=["lulc@264"], input_size=256, lr=3.535533905932738e-05, effective_batch_size=2048)
    """

    tokenizer = build_vqvae(
        model_type="vqvae",
        variant="terramind_v1_tokenizer_lulc",
        image_size=256,
        n_channels=10,
        encoder_type="vit_b_enc",
        decoder_type="vit_b_dec",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        quant_type="fsq",
        codebook_size="7-5-5-5-5",
        latent_dim=5,
        **kwargs
    )

    return tokenizer



@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v1_tokenizer_ndvi(**kwargs):
    """
    NDVI Tokenizer for TerraMind v1.

    model args: Namespace(patch_size=16, input_size_min=256, input_size_max=256, resolution_step=32, input_size_enc=None, encoder_type="vit_b_enc", decoder_type="unet_patched", post_mlp=True, encoder_ckpt=None, full_ckpt=None, freeze_enc=False, from_scratch=True, dec_transformer_dropout=0.2, quantizer_type="fsq", codebook_size="8-8-8-6-5", latent_dim=5, norm_codes=True, norm_latents=False, codebook_weight=1.0, quantizer_ema_decay=0.99, coef_ema_dead_code=32.0, code_replacement_policy="batch_random", commitment_weight=1.0, kmeans_init=False, num_train_timesteps=1000, prediction_type="sample", beta_schedule="linear", zero_terminal_snr=True, cls_free_guidance_dropout=0.0, masked_cfg=False, masked_cfg_low=0, masked_cfg_high=None, thresholding=True, loss_fn="mse", conditioning="concat", resolution_cond=False, eval_res_cond=None, batch_size=1, accum_grad=2, batch_size_eval=None, epochs=20, stop_after_epoch=250, save_ckpt_freq=1, opt="adamw", opt_eps=1e-08, opt_betas=[0.9, 0.99], clip_grad=1.0, skip_grad=None, momentum=0.9, weight_decay=0.05, weight_decay_end=0.05, blr=0.0064, warmup_lr=1e-06, min_lr=0.0, warmup_epochs=1, warmup_steps=-1, dtype="fp16", model_ema=True, model_ema_decay=0.9999, model_ema_force_cpu=False, model_ema_update_freq=1, hflip=0.5, domain="ndvi@264", mask_value=None, data_path="./data/TerraMesh/train", eval_data_path="./data/TerraMesh/val", imagenet_default_mean_and_std=False, standardize_surface_normals=False, min_crop_scale=0.8, cache_datasets=False, dataset_size=None, sample_sentinel_data=False, num_seasons=1, num_locations=64, dist_eval=True, step_eval=True, eval_noise_schedule="DDIMScheduler", num_eval_timesteps=50, input_size_eval=[256], num_eval_metrics_samples=5, eval_freq=2000, eval_metrics_freq=0, eval_image_log_freq=2000, num_logged_images=5, eval_only=False, no_inception=False, log_codebook_usage=True, output_dir="artifacts/runs/sm-ndvi-20-epochs", device="cuda", seed=0, resume="artifacts/runs/sm-ndvi-20-epochs/checkpoint_12.pth", auto_resume=True, start_epoch=13, num_workers=8, pin_mem=True, find_unused_params=True, log_wandb=True, wandb_project="terramind-tokenizers", wandb_entity="fasteo-ibm-jsc", wandb_run_name="2048_epochs_20/divae/rgb/NDVI-ViTB-UNetP4_16k_224-264", wandb_run_id=None, wandb_tags=[], show_user_warnings=False, world_size=16, local_rank=-1, dist_on_itp=False, dist_url="env://", run_name="tokenization/divae/rgb/NDVI-ViTB-UNetP4_16k_224-264", patch_size_dec=4, clip_sample=True, epoch_eval=False, config_path="/p/home/jusers/maurogiovanni1/juwels/4m4eo/cfgs/default/tokenization/divae/rgb/NDVI-ViTB-UNetP4_16k_224-264.yaml", rank=0, gpu=0, distributed=True, dist_backend="nccl", num_tasks=16, all_domains=["ndvi@264"], input_size=256, lr=0.0001414213562373095, effective_batch_size=2048)
    """

    tokenizer = build_vqvae(
        variant="terramind_v1_tokenizer_ndvi",
        image_size=256,
        n_channels=1,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v01_tokenizer_s2l2a(**kwargs):
    """
    S2L2A Tokenizer for TerraMind v0.1.

    This tokenizer was pre-trained using the SSL4EO-S12 v1.1 samples.
    """
    if kwargs.get("pretrained", False):
        if not os.getenv("HF_TOKEN", None):
            warnings.warn("TerraMind v0.1 models require a HF_TOKEN with access to model weights.")

    tokenizer = build_vqvae(
        variant="terramind_v01_tokenizer_s2l2a",
        image_size=256,
        n_channels=12,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v01_tokenizer_s1grd(**kwargs):
    """
    S1GRD Tokenizer for TerraMind v0.1.

    This tokenizer was pre-trained using the SSL4EO-S12 v1.1 samples.
    """
    if kwargs.get("pretrained", False):
        if not os.getenv("HF_TOKEN", None):
            warnings.warn("TerraMind v0.1 models require a HF_TOKEN with access to model weights.")

    tokenizer = build_vqvae(
        variant="terramind_v01_tokenizer_s1grd",
        image_size=256,
        n_channels=2,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v01_tokenizer_dem(**kwargs):
    """
    DEM Tokenizer for TerraMind v0.1.

    This tokenizer was pre-trained on the SSL4EO-S12 v1.1 locations.
    """
    if kwargs.get("pretrained", False):
        if not os.getenv("HF_TOKEN", None):
            warnings.warn("TerraMind v0.1 models require a HF_TOKEN with access to model weights.")

    tokenizer = build_vqvae(
        variant="terramind_v01_tokenizer_dem",
        image_size=256,
        n_channels=1,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer


@TERRATORCH_FULL_MODEL_REGISTRY.register
def terramind_v01_tokenizer_lulc(**kwargs):
    """
    LULC Tokenizer for TerraMind v0.1.

    This tokenizer was pre-trained on the SSL4EO-S12 v1.1 locations using the nine DynamicWorld classes,
    not with ESRI LULC as v1.0.
    """
    if kwargs.get("pretrained", False):
        if not os.getenv("HF_TOKEN", None):
            warnings.warn("TerraMind v0.1 models require a HF_TOKEN with access to model weights.")

    tokenizer = build_vqvae(
        model_type="vqvae",
        variant="terramind_v01_tokenizer_lulc",
        image_size=256,
        n_channels=9,
        encoder_type="vit_b_enc",
        decoder_type="vit_b_dec",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        quant_type="fsq",
        codebook_size="7-5-5-5-5",
        latent_dim=5,
        **kwargs
    )

    return tokenizer


def terramind_v01_caption_tokenizer(pretrained=True, tokenizer_file=None, *args, **kwargs):
    if not tokenizers_available:
        warnings.warn(f"Cannot import tokenizers. "
                      f"\nMake sure to install `pip install tokenizers`.")
        raise import_error_tokenizers

    if pretrained and tokenizer_file is None:
        tokenizer_file = hf_hub_download(
            repo_id=pretrained_weights["terramind_v01_caption_tokenizer"]["hf_hub_id"],
            filename=pretrained_weights["terramind_v01_caption_tokenizer"]["hf_hub_filename"]
        )

    return CaptionTokenizer(
        tokenizer_file=tokenizer_file,
        *args, **kwargs
    )


def terramind_v1_coords_tokenizer(pretrained=True, tokenizer_file=None, *args, **kwargs):
    if not tokenizers_available:
        warnings.warn(f"Cannot import tokenizers. "
                      f"\nMake sure to install `pip install tokenizers`.")
        raise import_error_tokenizers

    if pretrained and tokenizer_file is None:
        tokenizer_file = hf_hub_download(
            repo_id=pretrained_weights["terramind_v1_coords_tokenizer"]["hf_hub_id"],
            filename=pretrained_weights["terramind_v1_coords_tokenizer"]["hf_hub_filename"]
        )

    return CoordsTokenizer(
        tokenizer_file=tokenizer_file,
        *args, **kwargs
    )



from torch._tensor import Tensor
from granitewxc.datasets.merra2 import Merra2DownscaleDataset
from torchgeo.datamodules import NonGeoDataModule
from typing import Any, Callable, Optional
from granitewxc.datasets.merra2 import Merra2DownscaleDataset
from granitewxc.utils.config import ExperimentConfig
from torch.utils.data.dataloader import DataLoader
from torch._tensor import Tensor
from typing import Callable

class Merra2DownscaleNonGeoDataModule(NonGeoDataModule):

    def __init__(self,
                 data_path_surface: str,
                 data_path_vertical: str,
                 output_vars: list[str],
                 input_surface_vars: Optional[list[str]] = None,
                 input_static_surface_vars: Optional[list[str]] = None,
                 input_vertical_vars: Optional[list[str]] = None,
                 input_levels: Optional[list[float]] = None,
                 time_range: slice = None,
                 climatology_path_surface: Optional[str] = None,
                 climatology_path_vertical: Optional[str] = None,
                 transforms: list[Callable] = [],
                 n_input_timestamps = 1,
                 **kwargs: Any) -> None:
        
          super().__init__(Merra2DownscaleDataset,
                           time_range=time_range,
                           data_path_surface=data_path_surface,
                           data_path_vertical=data_path_vertical,
                           climatology_path_surface=climatology_path_surface,
                           climatology_path_vertical=climatology_path_vertical,
                           input_surface_vars=input_surface_vars,
                           input_static_surface_vars=input_static_surface_vars,
                           input_vertical_vars=input_vertical_vars,
                           input_levels=input_levels,
                           n_input_timestamps=n_input_timestamps,
                           output_vars=output_vars,
                           transforms=transforms,
                           **kwargs)

          self.aug = lambda x: x

    def _dataloader_factory(self, split: str) -> DataLoader[dict[str, Tensor]]:
        return super()._dataloader_factory(split)
    
    def setup(self, stage: str) -> None:

        if (stage == 'train'):
            self.train_dataset = self.dataset_class(  # type: ignore[call-arg]
                **self.kwargs
            ) 
        if (stage == 'val'):
            self.val_dataset = self.dataset_class(  # type: ignore[call-arg]
                **self.kwargs
            ) 
        if (stage == 'test'):
            self.test_dataset = self.dataset_class(  # type: ignore[call-arg]
                **self.kwargs
            ) 
        if (stage == 'predict'):
            self.predict_dataset = self.dataset_class(  # type: ignore[call-arg]
                **self.kwargs
            ) 

        return super().setup(stage)



import random
import torch
from torch.utils.data import Dataset
from torchvision import transforms
from PIL import Image
import numpy as np
import os
import matplotlib.pyplot as plt
import terratorch.visualization as ttv



class CopyPasteObjectDetectionDataset(Dataset):
    def __init__(
        self,
        base_dataset,              # yields (image, boxes)
        object_folder,             # folder with PNGs (RGBA)
        paste_prob=0.7,
        scale_range=(0.5, 1.5),
        max_objects=3,
        image_size=None            # (H, W) or None
    ):
        self.base_dataset = base_dataset
        self.object_paths = [
            os.path.join(object_folder, f)
            for f in os.listdir(object_folder)
            if f.lower().endswith(".png")
        ]
        assert len(self.object_paths) > 0

        self.paste_prob = paste_prob
        self.scale_range = scale_range
        self.max_objects = max_objects
        self.image_size = image_size

        self.to_tensor = transforms.ToTensor()

    def __len__(self):
        return len(self.base_dataset)

    def _load_object(self):
        obj = Image.open(random.choice(self.object_paths)).convert("RGBA")
        return obj

    def _paste_object(self, image, mask, obj):
        H, W = image.shape[-2:]

        scale = random.uniform(*self.scale_range)
        ow, oh = obj.size
        ow, oh = int(ow * scale), int(oh * scale)

        if ow >= W or oh >= H:
            return image, mask

        obj = obj.resize((ow, oh), Image.BILINEAR)
        ox = random.randint(0, W - ow)
        oy = random.randint(0, H - oh)

        obj_np = np.array(obj)
        rgb = torch.from_numpy(obj_np[..., :3]).permute(2, 0, 1) / 255.0
        alpha = torch.from_numpy(obj_np[..., 3] / 255.0)

        image[:, oy:oy+oh, ox:ox+ow] = (
            image[:, oy:oy+oh, ox:ox+ow] * (1 - alpha)
            + rgb * alpha
        )

        mask[oy:oy+oh, ox:ox+ow] |= (alpha > 0.5)

        bbox = torch.tensor([ox, oy, ox + ow, oy + oh], dtype=torch.float32)
        return image, mask, bbox

    def __getitem__(self, idx):
        item = self.base_dataset[idx]
        
        # Handle both dict and tuple returns
        if isinstance(item, dict):
            image = item['image']
            boxes = item['boxes']
            labels = item.get('labels', None)
        else:  # tuple format (image, boxes) or (image, boxes, labels)
            if len(item) == 2:
                image, boxes = item
                labels = None
            else:
                image, boxes, labels = item[:3]

        if isinstance(image, Image.Image):
            image = self.to_tensor(image)

        if self.image_size is not None:
            image = torch.nn.functional.interpolate(
                image.unsqueeze(0),
                size=self.image_size,
                mode="bilinear",
                align_corners=False
            ).squeeze(0)

        H, W = image.shape[-2:]
        mask = torch.zeros((H, W), dtype=torch.uint8)

        # mark original boxes
        for box in boxes:
            x1, y1, x2, y2 = box.int()
            mask[y1:y2, x1:x2] = 1

        if random.random() < self.paste_prob:
            n = random.randint(1, self.max_objects)
            for _ in range(n):
                obj = self._load_object()
                image, mask, bbox = self._paste_object(image, mask, obj)

                boxes = torch.cat([boxes, bbox.unsqueeze(0)], dim=0)

                # assuming class 1 for pasted objects
                if labels is not None:
                    labels = torch.cat([labels, torch.tensor([1], dtype=torch.long)], dim=0)

        return {
            "image": image.float(),
            "boxes": boxes,
            "mask": mask,
            "labels": labels
        }


    def plot(
        self,
        sample: dict[str, torch.Tensor],
        suptitle: str | None = None,
    ) -> plt.Figure:
        """Plot a detection sample using terratorch helpers (TensorBoard-safe)."""

        fig = ttv.plot_boxes_labels(
            image=sample["image"],
            boxes=sample["boxes"],
            show=False,          # REQUIRED for TensorBoard
        )

        if suptitle:
            fig.suptitle(suptitle)

        return fig



"""Code for Transformer from Phil Wangs vit-pytorch library.
Repository: https://github.com/lucidrains/vit-pytorch
"""

import torch
import torch.nn.functional as F
from einops import rearrange
from torch import nn


class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim):
        super().__init__()
        self.net = nn.Sequential(
            nn.LayerNorm(dim),
            nn.Linear(dim, hidden_dim),
            nn.GELU(),
            nn.Linear(hidden_dim, dim),
        )

    def forward(self, x):
        return self.net(x)


class Attention(nn.Module):
    def __init__(self, dim, heads=8, dim_head=64, fused_attn=True):
        super().__init__()
        inner_dim = dim_head * heads
        self.heads = heads
        self.scale = dim_head**-0.5
        self.norm = nn.LayerNorm(dim)
        self.fused_attn = fused_attn

        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)
        self.to_out = nn.Linear(inner_dim, dim, bias=False)

    def forward(self, x):
        x = self.norm(x)

        qkv = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(t, "b n (h d) -> b h n d", h=self.heads), qkv)

        if self.fused_attn:
            x = F.scaled_dot_product_attention(q, k, v, dropout_p=0.0)
        else:
            attn = torch.matmul(q, k.transpose(-1, -2)) * self.scale
            attn = attn.softmax(dim=-1)
            x = torch.matmul(attn, v)

        x = rearrange(x, "b h n d -> b n (h d)")
        return self.to_out(x)


class Transformer(nn.Module):
    def __init__(  # noqa: PLR0913
        self,
        dim,
        depth,
        heads,
        dim_head,
        mlp_dim,
        fused_attn,
    ):
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(
                nn.ModuleList(
                    [
                        Attention(
                            dim, heads=heads, dim_head=dim_head, fused_attn=fused_attn
                        ),
                        FeedForward(dim, mlp_dim),
                    ]
                )
            )

    def forward(self, x):
        for attn, ff in self.layers:
            x = attn(x) + x
            x = ff(x) + x
        return self.norm(x)



# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import warnings

import numpy as np
import torch
from torch import nn
from functools import partial
from einops import rearrange
import torch.nn.functional as F

from .encoder_embeddings import ImageEncoderEmbedding, ImageTokenEncoderEmbedding
from .decoder_embeddings import ImageTokenDecoderEmbedding
from .tm_utils import LayerNorm
from .modality_info import MODALITY_INFO
from .generate import (
    GenerationSampler,
    build_chained_generation_schedules,
    init_full_input_modality,
    init_empty_target_modality,
    init_conditioned_target_modality,
)
from .terramind import (
    TerraMind,
    build_tokenizer,
    build_modality_embeddings,
    build_output_modality_embeddings,
)


class TerraMindGeneration(nn.Module):
    """Modified TerraMind model for a "Thinking in Modalities" approach.

    Args:
        img_size (int): Input image size.
        modalities (list, dict, optional): List of modality keys and dicts, or dict with modality keys and values being
            ints (num_channels of modality) or nn.Module (patch embedding layer).
        output_modalities (list, optional): List of tokenized modalities used for the TiM approach. The TiM outputs are
            generated in the same order as specified in the given list. Defaults to [tbd].
        decoding_steps (list, int): Number of decoding steps for each TiM modality. Defaults to 1.
        temps (list, float): Sampling temperatures for each TiM modality. Defaults to 1.0.
        top_p (float): Top-p sampling threshold for TiM modalities. Ignored if set to 0.0. Defaults to 0.8.
        top_k (int): Top-k sampling threshold for TiM modalities. Ignored if set to 0. Defaults to 0.
        patch_size (int): Patch size.
        dim (int): Patch embedding dimension.
        encoder_depth (int): Depth of ViT / number of encoder blocks.
        num_heads (int): Number of attention heads in each ViT block.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool): If True, add a learnable bias to query, key, value.
        proj_bias (bool): If True, adds a bias to the attention out proj layer.
        mlp_bias (bool): If True, adds a learnable bias for the feedforward.
        num_register_tokens (int): Number of register tokens.
        act_layer (nn.Module): Activation layer.
        norm_layer (nn.Module): Normalization layer.
        gated_mlp (bool): If True, makes the feedforward gated (e.g., for SwiGLU)
        qk_norm (bool): If True, normalizes the query and keys (as in ViT-22B)
        tokenizer_dict (dict): Dictionary of tokenizers.
        pretrained (bool): If True, loads pretrained tokenizers.
    """

    def __init__(
            self,
            img_size: int = 224,
            modalities: list[str] | dict[str, int | nn.Module] | None = None,
            output_modalities: list[str] | None = None,
            decoding_steps: list[int] | int = 1,
            temps: list[float] | float = 1.0,
            top_p: float = 0.8,
            top_k: int = 0,
            timesteps: int = 50,
            patch_size: int = 16,
            dim: int = 768,
            encoder_depth: int = 12,
            decoder_depth: int = 12,
            num_heads: int = 12,
            mlp_ratio: float = 4.0,
            qkv_bias: bool = True,
            proj_bias: bool = True,
            mlp_bias: bool = True,
            num_register_tokens: int = 0,
            act_layer: nn.Module = nn.GELU,
            norm_layer: partial | nn.Module = partial(LayerNorm, eps=1e-6),
            gated_mlp: bool = False,
            qk_norm: bool = False,
            standardize: bool = False,
            offset: dict[str, float] | None = None,
            pretraining_mean: dict[str, list] | None = None,
            pretraining_std: dict[str, list] | None = None,
            tokenizer_dict: dict = None,
            pretrained: bool = False,
    ):
        super().__init__()

        if modalities is None or len(modalities) == 0:
            raise ValueError('Input modalities not provided.')
        elif isinstance(modalities, dict):
            modalities = [modalities]
        elif not isinstance(modalities, list):
            raise ValueError(f'Modalities must be a list of modality keys or a dict with embedding layers.')

        if output_modalities is None or len(output_modalities) == 0:
            raise ValueError('Output modalities not provided.')
        elif not isinstance(output_modalities, list):
            raise ValueError(f'Output modalities must be a list of modality keys.')

        # Parameters for generation schedule
        self.top_p = top_p
        self.top_k = top_k
        self.timesteps = timesteps
        self.decoding_steps = decoding_steps
        self.temps = temps

        # Init embeddings
        self.encoder_embeddings, mod_name_mapping = build_modality_embeddings(
            MODALITY_INFO, modalities, img_size=img_size, dim=dim, patch_size=patch_size)
        self.decoder_embeddings, decoder_name_mapping = build_output_modality_embeddings(
            MODALITY_INFO, output_modalities, img_size=img_size, dim=dim, patch_size=patch_size)
        self.output_modalities = list(decoder_name_mapping.values())  # Update output modality names
        self.mod_name_mapping = decoder_name_mapping | mod_name_mapping  # Merging dicts
        self.modalities = list(mod_name_mapping.keys())  # Further code expects list
        self.image_modalities = [key for key, value in self.encoder_embeddings.items()
             if isinstance(value, ImageEncoderEmbedding) or isinstance(value, ImageTokenEncoderEmbedding)]
        self.output_image_modalities = [key for key, value in self.decoder_embeddings.items()
                                        if isinstance(value, ImageTokenDecoderEmbedding)]
        self.output_mod_name_mapping = {v: k for k, v in decoder_name_mapping.items()}
        self.standardize = standardize

        if len(modalities) == 1 and self.mod_name_mapping[modalities[0]] == 'caption':
            # TODO Debug why text to image generations don't work.
            raise NotImplementedError(f"TerraMind v0.1 generations with only text input don't work yet.")

        if offset is not None:
            for mod, o in offset.items():
                # Add offset to mean values
                if mod not in self.mod_name_mapping:
                    warnings.warn(f"offset {mod} not defined in input or output modalities, ignoring offset.")
                    continue
                pretraining_mean[self.mod_name_mapping[mod]] = \
                    np.array(pretraining_mean[self.mod_name_mapping[mod]], dtype=float) + o

        self.pretraining_mean = {mod: torch.tensor(mean, dtype=torch.get_default_dtype())[None, :, None, None]
                                 for mod, mean in pretraining_mean.items()} if pretraining_mean else {}
        self.pretraining_std = {mod: torch.tensor(std, dtype=torch.get_default_dtype())[None, :, None, None]
                                for mod, std in pretraining_std.items()} if pretraining_std else {}

        # Build MAE model
        mae_model = TerraMind(
            encoder_embeddings=self.encoder_embeddings,
            decoder_embeddings=self.decoder_embeddings,
            modality_info=MODALITY_INFO,
            dim=dim,
            encoder_depth=encoder_depth,
            decoder_depth=decoder_depth,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            qkv_bias=qkv_bias,
            proj_bias=proj_bias,
            mlp_bias=mlp_bias,
            num_register_tokens=num_register_tokens,
            act_layer=act_layer,
            norm_layer=norm_layer,
            gated_mlp=gated_mlp,
            qk_norm=qk_norm,
        )

        self.sampler = GenerationSampler(mae_model)

        self.tokenizer = build_tokenizer(tokenizer_dict=tokenizer_dict,
                                         input_modalities=list(self.encoder_embeddings.keys()),
                                         output_modalities=list(self.decoder_embeddings.keys()),
                                         pretrained=pretrained)

    def to(self, device):
        super().to(device)
        # Move standardization values to device
        for mod, mean in self.pretraining_mean.items():
            self.pretraining_mean[mod] = mean.to(device)
        for mod, std in self.pretraining_std.items():
            self.pretraining_std[mod] = std.to(device)
        return self

    def forward(
            self,
            d: dict[str, torch.Tensor] | torch.Tensor | None = None,
            standardize: bool | None = None,
            timesteps: int = None,
            verbose: bool = False,
            **kwargs
    ) -> dict[str, torch.Tensor]:
        """
        Forward pass of the model.

        Args:
            d (dict, torch.Tensor): Dict of inputs or input tensor with shape (B, C, H, W)

            Alternatively, keyword arguments with modality=tensor.

        Returns:
            dict[str, torch.Tensor]: Dict of generated images
        """
        # Handle single image modality
        if not isinstance(d, dict):
            # Assuming first modality
            d = {self.modalities[0]: d}
        elif d is None or len(d) == 0:
            d = {}
            if len(kwargs) == 0:
                raise ValueError("No inputs provided.")

        # Add additional keyword args to input dict
        for key, value in kwargs.items():
            d[key] = value

        # Check for unknown modalities in input
        for mod in list(d.keys()):
            if mod not in self.mod_name_mapping:
                warnings.warn(f"Unknown input modality: {mod}. Ignoring input.")
                del d[mod]
        if len(d) == 0:
            raise ValueError("No valid inputs provided.")

        # Get batch size and device
        batch_size = len(list(d.values())[0])
        device = next(self.parameters()).device

        standardize = standardize if standardize is not None else self.standardize
        if standardize:
            for mod, value in d.items():
                if self.mod_name_mapping[mod] in self.pretraining_mean:
                    d[mod] = ((value - self.pretraining_mean[self.mod_name_mapping[mod]]) /
                              self.pretraining_std[self.mod_name_mapping[mod]])

        # Define the initial input
        input_dict = {}
        # Default values if no images are provided
        img_num_tokens, image_size = 196, (224, 224)
        for mod, value in d.items():
            if self.mod_name_mapping[mod] in self.image_modalities:
                input_shape = value.shape
            if self.mod_name_mapping[mod] in self.tokenizer:
                # Tokenize
                value = self.tokenizer[self.mod_name_mapping[mod]].encode(value, device)
                if not isinstance(value, dict):
                    value = value[-1]  # Select tokens from img tokenizer

            if self.mod_name_mapping[mod] in self.image_modalities:
                # Get image size and num tokens
                patch_size = self.encoder_embeddings[self.mod_name_mapping[mod]].patch_size
                img_num_tokens = int((input_shape[-1] / patch_size[-1]) * (input_shape[-2] / patch_size[-2]))
                image_size = (input_shape[-1], input_shape[-2])

                # Init raw image input masks
                value = {
                    "tensor": value,
                    "input_mask": torch.zeros(batch_size, img_num_tokens, dtype=torch.bool, device=device),
                    "target_mask": torch.ones(batch_size, img_num_tokens, dtype=torch.bool, device=device),
                    "decoder_attention_mask": torch.zeros(batch_size, img_num_tokens, dtype=torch.bool, device=device),
                }

            # Encode input and provide expected format
            input_dict[self.mod_name_mapping[mod]] = init_full_input_modality(
                value,
                MODALITY_INFO,
                self.mod_name_mapping[mod],
                device
            )

        # Initialize output modalities
        tokens_per_target = []
        autoregression_schemes = []
        token_decoding_schedules = []
        token_decoding_steps = []

        for mod in self.output_modalities:
            if mod in self.output_image_modalities:
                mod_num_tokens = img_num_tokens
                autoregression_schemes.append("roar")
                token_decoding_schedules.append("linear")
                token_decoding_steps.append(self.decoding_steps)
            else:
                # Get max length from modality info for sequence data
                mod_num_tokens = self.decoder_embeddings[mod].max_length
                autoregression_schemes.append("autoregressive")
                token_decoding_schedules.append(None)
                token_decoding_steps.append(None)
            tokens_per_target.append(mod_num_tokens)

            if mod in input_dict:
                # Modality in input and target
                input_dict[mod] = init_conditioned_target_modality(input_dict[mod], MODALITY_INFO, mod, mod_num_tokens)
            else:
                input_dict[mod] = init_empty_target_modality(MODALITY_INFO, mod, batch_size, mod_num_tokens, device)

        # Predict tokens of output modalities
        schedule = build_chained_generation_schedules(
            cond_domains=[self.mod_name_mapping[m] for m in d.keys()],
            target_domains=self.output_modalities,
            tokens_per_target=tokens_per_target,
            autoregression_schemes=autoregression_schemes,
            decoding_steps=token_decoding_steps,
            token_decoding_schedules=token_decoding_schedules,
            temps=[self.temps] * len(self.output_modalities),
            temp_schedules=["constant"] * len(self.output_modalities),
            cfg_scales=[1.0] * len(self.output_modalities),
            cfg_schedules=["constant"] * len(self.output_modalities),
            cfg_grow_conditioning=True,
        )

        out_dict = self.sampler.generate(
            input_dict,
            schedule,
            verbose=False,
            seed=random.randint(-(2 ** 31), 2 ** 31 - 1),
            top_p=self.top_p,
            top_k=self.top_k,
            num_tokens=sum(tokens_per_target),
            tokenizer=self.tokenizer
        )

        # TODO Vary timesteps based on codebook diversity
        timesteps = timesteps or self.timesteps
        out = {}
        for mod in self.output_modalities:
            tok = out_dict[mod]['tensor']
            if mod in self.output_image_modalities:
                patch_size = self.tokenizer[mod].patch_size
                tok = rearrange(tok, "b (nh nw) -> b nh nw",
                                nh=image_size[0] // patch_size, nw=image_size[1] // patch_size)

                out[self.output_mod_name_mapping[mod]] = self.tokenizer[mod].decode_tokens(
                    tok,
                    image_size=image_size,
                    timesteps=timesteps,
                    verbose=verbose
                )

            elif mod in self.output_modalities and mod in ['caption', 'coords']:
                out[self.output_mod_name_mapping[mod]] = self.tokenizer[mod].decode_text(out_dict)

        if standardize:
            for mod, value in out.items():
                if self.mod_name_mapping[mod] in self.pretraining_mean:
                    out[mod] = (value * self.pretraining_std[self.mod_name_mapping[mod]] +
                                self.pretraining_mean[self.mod_name_mapping[mod]])

        return out



import unittest
from unittest.mock import patch
import torch
from torch import nn

from terratorch.tasks.classification_tasks import ClassificationTask
from terratorch.models.model import ModelOutput
from terratorch.tasks.utils import _instantiate_from_path

class DummyModel(nn.Module):
    def __init__(self, num_classes=3):
        super().__init__()
        self.num_classes = num_classes

    def forward(self, x, **kwargs):
        batch_size = x.shape[0]
        return ModelOutput(output=torch.randn(batch_size, self.num_classes))

class DummyLoss(nn.Module):
    def __init__(self, factor=1.0):
        super().__init__()
        self.factor = factor

    def forward(self, preds, targets):
        targets_onehot = nn.functional.one_hot(targets, num_classes=preds.size(1)).float()
        return ((preds - targets_onehot) ** 2).mean() * self.factor

class TestClassificationTaskCustomLoss(unittest.TestCase):

    def setUp(self):
        self.model_args = {"num_classes": 3}

    @patch("terratorch.tasks.classification_tasks._instantiate_from_path")
    def test_custom_loss_str_path(self, mock_instantiate):
        # _instantiate_from_path soll unsere DummyLoss zurückgeben
        mock_instantiate.return_value = DummyLoss(factor=2.0)

        task = ClassificationTask(
            model_args=self.model_args,
            model=DummyModel(num_classes=3),
            loss="dummy.DummyLoss",
            custom_loss=True,
            custom_loss_kwargs={"factor": 2.0}
        )
        task.configure_losses()
        self.assertIsInstance(task.criterion, DummyLoss)
        self.assertEqual(task.criterion.factor, 2.0)

    def test_custom_loss_nn_module(self):
        # Loss wird direkt als nn.Module übergeben
        custom_loss = DummyLoss(factor=3.0)
        task = ClassificationTask(
            model_args=self.model_args,
            model=DummyModel(num_classes=3),
            loss=custom_loss,
            custom_loss=True
        )
        task.configure_losses()
        self.assertIsInstance(task.criterion, DummyLoss)
        self.assertEqual(task.criterion.factor, 3.0)

    def test_training_step_with_custom_loss(self):
        custom_loss = DummyLoss(factor=1.5)
        task = ClassificationTask(
            model_args=self.model_args,
            model=DummyModel(num_classes=3),
            loss=custom_loss,
            custom_loss=True
        )
        task.configure_losses()
        # Dummy-Metrics konfigurieren
        task.configure_metrics()

        batch = {
            "image": torch.randn(2, 3, 64, 64),
            "label": torch.randint(0, 3, (2,))
        }
        loss = task.training_step(batch, batch_idx=0)
        self.assertIsInstance(loss, torch.Tensor)

class TestInstantiateFromPath(unittest.TestCase):

    def test_instantiate_class_directly(self):
        module_path = "tests.test_classification_tasks.DummyLoss" 
        obj = _instantiate_from_path(module_path, factor=2.0)
        self.assertIsInstance(obj, DummyLoss)
        self.assertEqual(obj.factor, 2.0)

    def test_instantiate_with_wrong_path(self):
        with self.assertRaises(ImportError):
            _instantiate_from_path("nonexistent.module.Class")

    def test_instantiate_with_wrong_class_name(self):
        import torch
        with self.assertRaises(AttributeError):
            _instantiate_from_path("torch.nn.NonExistentClass")

if __name__ == "__main__":
    unittest.main()



"""Comprehensive tests for Clay v1.5 backbone components.

Tests cover FeedForward, Attention, and Transformer modules with various configurations
to ensure maximum code coverage.
"""

import gc

import pytest
import torch
from torch import nn

from terratorch.models.backbones.clay_v15.backbone import (
    Attention,
    FeedForward,
    Transformer,
)


class TestFeedForward:
    """Test suite for FeedForward module."""

    @pytest.fixture
    def sample_input(self):
        """Create sample input tensor for testing."""
        # Batch size 2, sequence length 10, dimension 64
        return torch.randn(2, 10, 64)

    def test_feedforward_initialization(self):
        """Test FeedForward module initialization."""
        dim = 64
        hidden_dim = 256
        ff = FeedForward(dim=dim, hidden_dim=hidden_dim)

        assert isinstance(ff.net, nn.Sequential)
        assert len(ff.net) == 4
        assert isinstance(ff.net[0], nn.LayerNorm)
        assert isinstance(ff.net[1], nn.Linear)
        assert isinstance(ff.net[2], nn.GELU)
        assert isinstance(ff.net[3], nn.Linear)

        # Check layer dimensions
        assert ff.net[1].in_features == dim
        assert ff.net[1].out_features == hidden_dim
        assert ff.net[3].in_features == hidden_dim
        assert ff.net[3].out_features == dim

    def test_feedforward_forward(self, sample_input):
        """Test FeedForward forward pass."""
        dim = 64
        hidden_dim = 256
        ff = FeedForward(dim=dim, hidden_dim=hidden_dim)

        output = ff(sample_input)

        # Check output shape matches input shape
        assert output.shape == sample_input.shape
        # Check output is different from input (transformation occurred)
        assert not torch.allclose(output, sample_input)

    def test_feedforward_with_different_dimensions(self):
        """Test FeedForward with various dimension configurations."""
        test_cases = [
            (32, 128),
            (128, 512),
            (256, 1024),
            (512, 2048),
        ]

        for dim, hidden_dim in test_cases:
            ff = FeedForward(dim=dim, hidden_dim=hidden_dim)
            input_tensor = torch.randn(2, 10, dim)
            output = ff(input_tensor)
            assert output.shape == input_tensor.shape

    def test_feedforward_gradient_flow(self, sample_input):
        """Test that gradients flow correctly through FeedForward."""
        dim = 64
        hidden_dim = 256
        ff = FeedForward(dim=dim, hidden_dim=hidden_dim)

        sample_input.requires_grad = True
        output = ff(sample_input)
        loss = output.sum()
        loss.backward()

        # Check that gradients are computed
        assert sample_input.grad is not None
        assert not torch.all(sample_input.grad == 0)

    def test_feedforward_batch_independence(self):
        """Test that FeedForward processes batch samples independently."""
        dim = 64
        hidden_dim = 256
        ff = FeedForward(dim=dim, hidden_dim=hidden_dim)

        # Create input with two different samples
        input1 = torch.randn(1, 10, dim)
        input2 = torch.randn(1, 10, dim)
        batched_input = torch.cat([input1, input2], dim=0)

        # Process separately and together
        output1 = ff(input1)
        output2 = ff(input2)
        batched_output = ff(batched_input)

        # Results should be identical
        assert torch.allclose(batched_output[0], output1[0], atol=1e-6)
        assert torch.allclose(batched_output[1], output2[0], atol=1e-6)


class TestAttention:
    """Test suite for Attention module."""

    @pytest.fixture
    def sample_input(self):
        """Create sample input tensor for testing."""
        # Batch size 2, sequence length 10, dimension 64
        return torch.randn(2, 10, 64)

    def test_attention_initialization(self):
        """Test Attention module initialization."""
        dim = 64
        heads = 8
        dim_head = 64
        fused_attn = True

        attn = Attention(dim=dim, heads=heads, dim_head=dim_head, fused_attn=fused_attn)

        assert attn.heads == heads
        assert attn.scale == dim_head**-0.5
        assert attn.fused_attn == fused_attn
        assert isinstance(attn.norm, nn.LayerNorm)
        assert isinstance(attn.to_qkv, nn.Linear)
        assert isinstance(attn.to_out, nn.Linear)

        # Check dimensions
        inner_dim = dim_head * heads
        assert attn.to_qkv.in_features == dim
        assert attn.to_qkv.out_features == inner_dim * 3
        assert attn.to_out.in_features == inner_dim
        assert attn.to_out.out_features == dim

    def test_attention_default_parameters(self):
        """Test Attention with default parameters."""
        dim = 64
        attn = Attention(dim=dim)

        # Check defaults
        assert attn.heads == 8
        assert attn.scale == 64**-0.5
        assert attn.fused_attn == True

    def test_attention_forward_fused(self, sample_input):
        """Test Attention forward pass with fused attention."""
        dim = 64
        attn = Attention(dim=dim, heads=8, dim_head=64, fused_attn=True)

        output = attn(sample_input)

        # Check output shape matches input shape
        assert output.shape == sample_input.shape

    def test_attention_forward_unfused(self, sample_input):
        """Test Attention forward pass without fused attention."""
        dim = 64
        attn = Attention(dim=dim, heads=8, dim_head=64, fused_attn=False)

        output = attn(sample_input)

        # Check output shape matches input shape
        assert output.shape == sample_input.shape

    def test_attention_fused_vs_unfused(self, sample_input):
        """Test that fused and unfused attention produce similar results."""
        dim = 64
        heads = 4
        dim_head = 32

        # Create two attention modules with same initialization
        torch.manual_seed(42)
        attn_fused = Attention(dim=dim, heads=heads, dim_head=dim_head, fused_attn=True)

        torch.manual_seed(42)
        attn_unfused = Attention(dim=dim, heads=heads, dim_head=dim_head, fused_attn=False)

        # Set to eval mode for deterministic behavior
        attn_fused.eval()
        attn_unfused.eval()

        with torch.no_grad():
            output_fused = attn_fused(sample_input)
            output_unfused = attn_unfused(sample_input)

        # Results should be very similar (allowing for numerical differences)
        assert torch.allclose(output_fused, output_unfused, atol=1e-5, rtol=1e-4)

    def test_attention_with_different_heads(self):
        """Test Attention with various number of heads."""
        dim = 64
        test_cases = [1, 2, 4, 8, 16]

        for heads in test_cases:
            attn = Attention(dim=dim, heads=heads, dim_head=32)
            input_tensor = torch.randn(2, 10, dim)
            output = attn(input_tensor)
            assert output.shape == input_tensor.shape

    def test_attention_with_different_dim_head(self):
        """Test Attention with various dimension per head."""
        dim = 128
        test_cases = [16, 32, 64, 128]

        for dim_head in test_cases:
            attn = Attention(dim=dim, heads=4, dim_head=dim_head)
            input_tensor = torch.randn(2, 10, dim)
            output = attn(input_tensor)
            assert output.shape == input_tensor.shape

    def test_attention_gradient_flow(self, sample_input):
        """Test that gradients flow correctly through Attention."""
        dim = 64
        attn = Attention(dim=dim, heads=8, dim_head=64)

        sample_input.requires_grad = True
        output = attn(sample_input)
        loss = output.sum()
        loss.backward()

        # Check that gradients are computed
        assert sample_input.grad is not None
        assert not torch.all(sample_input.grad == 0)

    def test_attention_with_different_sequence_lengths(self):
        """Test Attention with various sequence lengths."""
        dim = 64
        attn = Attention(dim=dim, heads=8, dim_head=64)

        test_lengths = [1, 5, 10, 50, 100]
        for seq_len in test_lengths:
            input_tensor = torch.randn(2, seq_len, dim)
            output = attn(input_tensor)
            assert output.shape == input_tensor.shape

    def test_attention_single_element(self):
        """Test Attention with single element in sequence."""
        dim = 64
        attn = Attention(dim=dim, heads=8, dim_head=64)

        input_tensor = torch.randn(1, 1, dim)
        output = attn(input_tensor)
        assert output.shape == input_tensor.shape

    def test_attention_batch_independence(self):
        """Test that Attention processes batch samples independently."""
        dim = 64
        attn = Attention(dim=dim, heads=4, dim_head=32, fused_attn=False)

        # Create input with two different samples
        input1 = torch.randn(1, 10, dim)
        input2 = torch.randn(1, 10, dim)
        batched_input = torch.cat([input1, input2], dim=0)

        # Process separately and together
        output1 = attn(input1)
        output2 = attn(input2)
        batched_output = attn(batched_input)

        # Results should be identical
        assert torch.allclose(batched_output[0], output1[0], atol=1e-6)
        assert torch.allclose(batched_output[1], output2[0], atol=1e-6)


class TestTransformer:
    """Test suite for Transformer module."""

    @pytest.fixture
    def sample_input(self):
        """Create sample input tensor for testing."""
        # Batch size 2, sequence length 10, dimension 64
        return torch.randn(2, 10, 64)

    def test_transformer_initialization(self):
        """Test Transformer module initialization."""
        dim = 64
        depth = 4
        heads = 8
        dim_head = 64
        mlp_dim = 256
        fused_attn = True

        transformer = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=fused_attn,
        )

        assert isinstance(transformer.norm, nn.LayerNorm)
        assert len(transformer.layers) == depth

        # Check each layer has attention and feedforward
        for layer in transformer.layers:
            assert isinstance(layer, nn.ModuleList)
            assert len(layer) == 2
            assert isinstance(layer[0], Attention)
            assert isinstance(layer[1], FeedForward)

    def test_transformer_forward(self, sample_input):
        """Test Transformer forward pass."""
        dim = 64
        depth = 2
        heads = 4
        dim_head = 32
        mlp_dim = 128
        fused_attn = True

        transformer = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=fused_attn,
        )

        output = transformer(sample_input)

        # Check output shape matches input shape
        assert output.shape == sample_input.shape

    def test_transformer_with_various_depths(self):
        """Test Transformer with different depths."""
        dim = 64
        heads = 4
        dim_head = 32
        mlp_dim = 128
        test_depths = [1, 2, 4, 8, 12]

        for depth in test_depths:
            transformer = Transformer(
                dim=dim,
                depth=depth,
                heads=heads,
                dim_head=dim_head,
                mlp_dim=mlp_dim,
                fused_attn=True,
            )
            input_tensor = torch.randn(2, 10, dim)
            output = transformer(input_tensor)
            assert output.shape == input_tensor.shape
            assert len(transformer.layers) == depth

    def test_transformer_fused_vs_unfused(self, sample_input):
        """Test Transformer with fused and unfused attention."""
        dim = 64
        depth = 2
        heads = 4
        dim_head = 32
        mlp_dim = 128

        # Create two transformers with same initialization
        torch.manual_seed(42)
        transformer_fused = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=True,
        )

        torch.manual_seed(42)
        transformer_unfused = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=False,
        )

        # Set to eval mode for deterministic behavior
        transformer_fused.eval()
        transformer_unfused.eval()

        with torch.no_grad():
            output_fused = transformer_fused(sample_input)
            output_unfused = transformer_unfused(sample_input)

        # Results should be very similar
        assert torch.allclose(output_fused, output_unfused, atol=1e-5, rtol=1e-4)

    def test_transformer_gradient_flow(self, sample_input):
        """Test that gradients flow correctly through Transformer."""
        dim = 64
        depth = 2
        heads = 4
        dim_head = 32
        mlp_dim = 128

        transformer = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=True,
        )

        sample_input.requires_grad = True
        output = transformer(sample_input)
        loss = output.sum()
        loss.backward()

        # Check that gradients are computed
        assert sample_input.grad is not None
        assert not torch.all(sample_input.grad == 0)

    def test_transformer_residual_connections(self):
        """Test that residual connections work correctly."""
        dim = 64
        depth = 1
        heads = 4
        dim_head = 32
        mlp_dim = 128

        transformer = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=True,
        )

        # Use identity-like initialization to observe residuals
        input_tensor = torch.randn(2, 10, dim)
        output = transformer(input_tensor)

        # Output should be different due to transformations but have similar magnitude
        assert output.shape == input_tensor.shape
        # The residual connections should prevent the output from being zero
        assert not torch.all(output == 0)

    def test_transformer_with_different_dimensions(self):
        """Test Transformer with various dimension configurations."""
        test_configs = [
            (32, 2, 4, 16, 64),
            (64, 4, 8, 32, 128),
            (128, 6, 8, 64, 256),
            (192, 6, 4, 48, 384),
        ]

        for dim, depth, heads, dim_head, mlp_dim in test_configs:
            transformer = Transformer(
                dim=dim,
                depth=depth,
                heads=heads,
                dim_head=dim_head,
                mlp_dim=mlp_dim,
                fused_attn=True,
            )
            input_tensor = torch.randn(2, 10, dim)
            output = transformer(input_tensor)
            assert output.shape == input_tensor.shape

    def test_transformer_with_different_sequence_lengths(self):
        """Test Transformer with various sequence lengths."""
        dim = 64
        depth = 2
        heads = 4
        dim_head = 32
        mlp_dim = 128

        transformer = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=True,
        )

        test_lengths = [1, 5, 10, 50, 100, 196]
        for seq_len in test_lengths:
            input_tensor = torch.randn(2, seq_len, dim)
            output = transformer(input_tensor)
            assert output.shape == input_tensor.shape

    def test_transformer_batch_independence(self):
        """Test that Transformer processes batch samples independently."""
        dim = 64
        depth = 2
        heads = 4
        dim_head = 32
        mlp_dim = 128

        transformer = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=False,
        )

        # Create input with two different samples
        input1 = torch.randn(1, 10, dim)
        input2 = torch.randn(1, 10, dim)
        batched_input = torch.cat([input1, input2], dim=0)

        # Process separately and together
        output1 = transformer(input1)
        output2 = transformer(input2)
        batched_output = transformer(batched_input)

        # Results should be identical
        assert torch.allclose(batched_output[0], output1[0], atol=1e-6)
        assert torch.allclose(batched_output[1], output2[0], atol=1e-6)

    def test_transformer_normalization(self, sample_input):
        """Test that final layer normalization is applied."""
        dim = 64
        depth = 2
        heads = 4
        dim_head = 32
        mlp_dim = 128

        transformer = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=True,
        )

        output = transformer(sample_input)

        # Output should have reasonable statistics due to normalization
        # Mean should be close to 0 and std close to 1 for each feature
        assert output.shape == sample_input.shape
        assert torch.isfinite(output).all()

    def test_transformer_eval_mode(self, sample_input):
        """Test Transformer in evaluation mode."""
        dim = 64
        depth = 2
        heads = 4
        dim_head = 32
        mlp_dim = 128

        transformer = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=True,
        )

        transformer.eval()
        with torch.no_grad():
            output1 = transformer(sample_input)
            output2 = transformer(sample_input)

        # In eval mode with same input, should get identical results
        assert torch.allclose(output1, output2)

    def test_transformer_train_mode(self, sample_input):
        """Test Transformer in training mode."""
        dim = 64
        depth = 2
        heads = 4
        dim_head = 32
        mlp_dim = 128

        transformer = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=True,
        )

        transformer.train()
        output = transformer(sample_input)

        # Should produce valid output
        assert output.shape == sample_input.shape
        assert torch.isfinite(output).all()

    def test_transformer_memory_efficiency(self):
        """Test that Transformer doesn't cause memory issues."""
        dim = 64
        depth = 4
        heads = 8
        dim_head = 32
        mlp_dim = 128

        transformer = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=True,
        )

        input_tensor = torch.randn(4, 196, dim)
        output = transformer(input_tensor)

        # Clean up
        del transformer, input_tensor, output
        gc.collect()


class TestIntegration:
    """Integration tests combining multiple components."""

    def test_stacked_transformers(self):
        """Test stacking multiple Transformer modules."""
        dim = 64
        depth = 2
        heads = 4
        dim_head = 32
        mlp_dim = 128

        transformer1 = Transformer(
            dim=dim, depth=depth, heads=heads, dim_head=dim_head,
            mlp_dim=mlp_dim, fused_attn=True
        )
        transformer2 = Transformer(
            dim=dim, depth=depth, heads=heads, dim_head=dim_head,
            mlp_dim=mlp_dim, fused_attn=True
        )

        input_tensor = torch.randn(2, 10, dim)
        intermediate = transformer1(input_tensor)
        output = transformer2(intermediate)

        assert output.shape == input_tensor.shape

    def test_large_model_config(self):
        """Test configuration similar to large models."""
        # Configuration similar to ViT-Large
        dim = 192
        depth = 6
        heads = 4
        dim_head = 48
        mlp_dim = 384

        transformer = Transformer(
            dim=dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            fused_attn=True,
        )

        # Typical image patch embedding (14x14 patches + 1 cls token)
        input_tensor = torch.randn(2, 197, dim)
        output = transformer(input_tensor)

        assert output.shape == input_tensor.shape
        gc.collect()

    def test_pipeline_with_all_components(self):
        """Test a full pipeline using all components."""
        dim = 64
        heads = 4
        dim_head = 32
        mlp_dim = 128

        # Create individual components
        attention = Attention(dim=dim, heads=heads, dim_head=dim_head, fused_attn=True)
        feedforward = FeedForward(dim=dim, hidden_dim=mlp_dim)
        transformer = Transformer(
            dim=dim, depth=2, heads=heads, dim_head=dim_head,
            mlp_dim=mlp_dim, fused_attn=True
        )

        input_tensor = torch.randn(2, 10, dim)

        # Pass through components
        attn_out = attention(input_tensor)
        ff_out = feedforward(attn_out)
        final_out = transformer(ff_out)

        assert final_out.shape == input_tensor.shape
        gc.collect()

    def test_mixed_precision_compatibility(self):
        """Test that modules work with mixed precision."""
        dim = 64
        depth = 2
        heads = 4
        dim_head = 32
        mlp_dim = 128

        transformer = Transformer(
            dim=dim, depth=depth, heads=heads, dim_head=dim_head,
            mlp_dim=mlp_dim, fused_attn=True
        )

        # Test with float16
        input_tensor = torch.randn(2, 10, dim, dtype=torch.float16)
        transformer = transformer.half()

        output = transformer(input_tensor)
        assert output.dtype == torch.float16
        assert output.shape == input_tensor.shape

        gc.collect()

    def test_device_placement(self):
        """Test that modules can be placed on different devices."""
        dim = 64
        depth = 2
        heads = 4
        dim_head = 32
        mlp_dim = 128

        transformer = Transformer(
            dim=dim, depth=depth, heads=heads, dim_head=dim_head,
            mlp_dim=mlp_dim, fused_attn=True
        )

        # Test on CPU
        input_cpu = torch.randn(2, 10, dim)
        output_cpu = transformer(input_cpu)
        assert output_cpu.device == torch.device('cpu')

        # Test on CUDA if available
        if torch.cuda.is_available():
            transformer = transformer.cuda()
            input_cuda = torch.randn(2, 10, dim, device='cuda')
            output_cuda = transformer(input_cuda)
            assert output_cuda.device.type == 'cuda'

        gc.collect()


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ---
#
# This project includes code adapted from the original work by EPFL and Apple Inc.,
# licensed under the Apache License, Version 2.0.
# Source: https://github.com/apple/ml-4m/

"""
Helpers to train with 16-bit precision.
"""

import numpy as np
import torch as th
import torch.nn as nn
from torch._utils import _flatten_dense_tensors, _unflatten_dense_tensors

INITIAL_LOG_LOSS_SCALE = 20.0


def convert_module_to_f16(l):
    """
    Convert primitive modules to float16.
    """
    if isinstance(l, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
        l.weight.data = l.weight.data.half()
        if l.bias is not None:
            l.bias.data = l.bias.data.half()


def convert_module_to_f32(l):
    """
    Convert primitive modules to float32, undoing convert_module_to_f16().
    """
    if isinstance(l, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
        l.weight.data = l.weight.data.float()
        if l.bias is not None:
            l.bias.data = l.bias.data.float()

def convert_module_to_bf16(l):
    """
    Convert primitive modules to bfloat16, undoing convert_module_to_f16().
    """
    if isinstance(l, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
        l.weight.data = l.weight.data.bfloat16()
        if l.bias is not None:
            l.bias.data = l.bias.data.bfloat16()


def make_master_params(param_groups_and_shapes):
    """
    Copy model parameters into a (differently-shaped) list of full-precision
    parameters.
    """
    master_params = []
    for param_group, shape in param_groups_and_shapes:
        master_param = nn.Parameter(
            _flatten_dense_tensors(
                [param.detach().float() for (_, param) in param_group]
            ).view(shape)
        )
        master_param.requires_grad = True
        master_params.append(master_param)
    return master_params


def model_grads_to_master_grads(param_groups_and_shapes, master_params):
    """
    Copy the gradients from the model parameters into the master parameters
    from make_master_params().
    """
    for master_param, (param_group, shape) in zip(
        master_params, param_groups_and_shapes
    ):
        master_param.grad = _flatten_dense_tensors(
            [param_grad_or_zeros(param) for (_, param) in param_group]
        ).view(shape)


def master_params_to_model_params(param_groups_and_shapes, master_params):
    """
    Copy the master parameter data back into the model parameters.
    """
    # Without copying to a list, if a generator is passed, this will
    # silently not copy any parameters.
    for master_param, (param_group, _) in zip(master_params, param_groups_and_shapes):
        for (_, param), unflat_master_param in zip(
            param_group, unflatten_master_params(param_group, master_param.view(-1))
        ):
            param.detach().copy_(unflat_master_param)


def unflatten_master_params(param_group, master_param):
    return _unflatten_dense_tensors(master_param, [param for (_, param) in param_group])


def get_param_groups_and_shapes(named_model_params):
    named_model_params = list(named_model_params)
    scalar_vector_named_params = (
        [(n, p) for (n, p) in named_model_params if p.ndim <= 1],
        (-1),
    )
    matrix_named_params = (
        [(n, p) for (n, p) in named_model_params if p.ndim > 1],
        (1, -1),
    )
    return [scalar_vector_named_params, matrix_named_params]


def master_params_to_state_dict(
    model, param_groups_and_shapes, master_params, use_fp16
):
    if use_fp16:
        state_dict = model.state_dict()
        for master_param, (param_group, _) in zip(
            master_params, param_groups_and_shapes
        ):
            for (name, _), unflat_master_param in zip(
                param_group, unflatten_master_params(param_group, master_param.view(-1))
            ):
                assert name in state_dict
                state_dict[name] = unflat_master_param
    else:
        state_dict = model.state_dict()
        for i, (name, _value) in enumerate(model.named_parameters()):
            assert name in state_dict
            state_dict[name] = master_params[i]
    return state_dict


def state_dict_to_master_params(model, state_dict, use_fp16):
    if use_fp16:
        named_model_params = [
            (name, state_dict[name]) for name, _ in model.named_parameters()
        ]
        param_groups_and_shapes = get_param_groups_and_shapes(named_model_params)
        master_params = make_master_params(param_groups_and_shapes)
    else:
        master_params = [state_dict[name] for name, _ in model.named_parameters()]
    return master_params


def zero_master_grads(master_params):
    for param in master_params:
        param.grad = None


def zero_grad(model_params):
    for param in model_params:
        # Taken from https://pytorch.org/docs/stable/_modules/torch/optim/optimizer.html#Optimizer.add_param_group
        if param.grad is not None:
            param.grad.detach_()
            param.grad.zero_()


def param_grad_or_zeros(param):
    if param.grad is not None:
        return param.grad.data.detach()
    else:
        return th.zeros_like(param)


class MixedPrecisionTrainer:
    def __init__(
        self,
        *,
        model,
        use_fp16=False,
        fp16_scale_growth=1e-3,
        initial_lg_loss_scale=INITIAL_LOG_LOSS_SCALE,
    ):
        self.model = model
        self.use_fp16 = use_fp16
        self.fp16_scale_growth = fp16_scale_growth

        self.model_params = list(self.model.parameters())
        self.master_params = self.model_params
        self.param_groups_and_shapes = None
        self.lg_loss_scale = initial_lg_loss_scale

        if self.use_fp16:
            self.param_groups_and_shapes = get_param_groups_and_shapes(
                self.model.named_parameters()
            )
            self.master_params = make_master_params(self.param_groups_and_shapes)
            self.model.convert_to_fp16()

    def zero_grad(self):
        zero_grad(self.model_params)

    def backward(self, loss: th.Tensor):
        if self.use_fp16:
            loss_scale = 2 ** self.lg_loss_scale
            (loss * loss_scale).backward()
        else:
            loss.backward()

    def optimize(self, opt: th.optim.Optimizer):
        if self.use_fp16:
            return self._optimize_fp16(opt)
        else:
            return self._optimize_normal(opt)

    def _optimize_fp16(self, opt: th.optim.Optimizer):
        model_grads_to_master_grads(self.param_groups_and_shapes, self.master_params)
        grad_norm, param_norm = self._compute_norms(grad_scale=2 ** self.lg_loss_scale)
        if check_overflow(grad_norm):
            self.lg_loss_scale -= 1
            zero_master_grads(self.master_params)
            return False

        for p in self.master_params:
            p.grad.mul_(1.0 / (2 ** self.lg_loss_scale))
        opt.step()
        zero_master_grads(self.master_params)
        master_params_to_model_params(self.param_groups_and_shapes, self.master_params)
        self.lg_loss_scale += self.fp16_scale_growth
        return True

    def _optimize_normal(self, opt: th.optim.Optimizer):
        grad_norm, param_norm = self._compute_norms()
        opt.step()
        return True

    def _compute_norms(self, grad_scale=1.0):
        grad_norm = 0.0
        param_norm = 0.0
        for p in self.master_params:
            with th.no_grad():
                param_norm += th.norm(p, p=2, dtype=th.float).item() ** 2
                if p.grad is not None:
                    grad_norm += th.norm(p.grad, p=2, dtype=th.float).item() ** 2
        return np.sqrt(grad_norm) / grad_scale, np.sqrt(param_norm)

    def master_params_to_state_dict(self, master_params):
        return master_params_to_state_dict(
            self.model, self.param_groups_and_shapes, master_params, self.use_fp16
        )

    def state_dict_to_master_params(self, state_dict):
        return state_dict_to_master_params(self.model, state_dict, self.use_fp16)


def check_overflow(value):
    return (value == float("inf")) or (value == -float("inf")) or (value != value)


import json
import os
import warnings
from datetime import datetime, timezone

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from torchgeo.datasets import NonGeoDataset

from terratorch.datasets.utils import pad_dates_numpy, pad_numpy


class PASTIS(NonGeoDataset):
    """ "
    Pytorch Dataset class to load samples from the [PASTIS](https://github.com/VSainteuf/pastis-benchmark) dataset,
    for semantic and panoptic segmentation.
    """

    def __init__(
        self,
        data_root,
        norm=True,  # noqa: FBT002
        target="semantic",
        folds=None,
        reference_date="2018-09-01",
        date_interval=(-200, 600),
        class_mapping=None,
        transform=None,
        truncate_image=None,
        pad_image=None,
        satellites=["S2"],  # noqa: B006
    ):
        """

        Args:
            data_root (str): Path to the dataset.
            norm (bool): If true, images are standardised using pre-computed
                channel-wise means and standard deviations.
            reference_date (str, Format : 'YYYY-MM-DD'): Defines the reference date
                based on which all observation dates are expressed. Along with the image
                time series and the target tensor, this dataloader yields the sequence
                of observation dates (in terms of number of days since the reference
                date). This sequence of dates is used for instance for the positional
                encoding in attention based approaches.
            target (str): 'semantic' or 'instance'. Defines which type of target is
                returned by the dataloader.
                * If 'semantic' the target tensor is a tensor containing the class of
                each pixel.
                * If 'instance' the target tensor is the concatenation of several
                signals, necessary to train the Parcel-as-Points module:
                    - the centerness heatmap,
                    - the instance ids,
                    - the voronoi partitioning of the patch with regards to the parcels'
                    centers,
                    - the (height, width) size of each parcel,
                    - the semantic label of each parcel,
                    - the semantic label of each pixel.
            folds (list, optional): List of ints specifying which of the 5 official
                folds to load. By default (when None is specified), all folds are loaded.
            class_mapping (dict, optional): A dictionary to define a mapping between the
                default 18 class nomenclature and another class grouping. If not provided,
                the default class mapping is used.
            transform (callable, optional): A transform to apply to the loaded data
                (images, dates, and masks). By default, no transformation is applied.
            truncate_image (int, optional): Truncate the time dimension of the image to
                a specified number of timesteps. If None, no truncation is performed.
            pad_image (int, optional): Pad the time dimension of the image to a specified
                number of timesteps. If None, no padding is applied.
            satellites (list): Defines the satellites to use. If you are using PASTIS-R, you
                have access to Sentinel-2 imagery and Sentinel-1 observations in Ascending
                and Descending orbits, respectively S2, S1A, and S1D. For example, use
                satellites=['S2', 'S1A'] for Sentinel-2 + Sentinel-1 ascending time series,
                or satellites=['S2', 'S1A', 'S1D'] to retrieve all time series. If you are using
                PASTIS, only S2 observations are available.
        """
        if target not in ["semantic", "instance"]:
            msg = f"Target '{target}' not recognized. Use 'semantic', or 'instance'."
            raise ValueError(msg)
        valid_satellites = {"S2", "S1A", "S1D"}
        for sat in satellites:
            if sat not in valid_satellites:
                msg = f"Satellite '{sat}' not recognized. Valid options are {valid_satellites}."
                raise ValueError(msg)

        super().__init__()
        self.data_root = data_root
        self.norm = norm
        self.reference_date = datetime(*map(int, reference_date.split("-")), tzinfo=timezone.utc)
        self.class_mapping = np.vectorize(lambda x: class_mapping[x]) if class_mapping is not None else class_mapping
        self.target = target
        self.satellites = satellites
        self.transform = transform
        self.truncate_image = truncate_image
        self.pad_image = pad_image
        # loads patches metadata
        self.meta_patch = gpd.read_file(os.path.join(data_root, "metadata.geojson"))
        self.meta_patch.index = self.meta_patch["ID_PATCH"].astype(int)
        self.meta_patch.sort_index(inplace=True)
        # stores table for each satalite date
        self.date_tables = {s: None for s in satellites}
        # date interval used in the PASTIS benchmark paper.
        date_interval_begin, date_interval_end = date_interval
        self.date_range = np.array(range(date_interval_begin, date_interval_end))
        for s in satellites:
            # maps patches to its observation dates
            dates = self.meta_patch[f"dates-{s}"]
            date_table = pd.DataFrame(index=self.meta_patch.index, columns=self.date_range, dtype=int)
            for pid, date_seq in dates.items():
                if type(date_seq) is str:
                    date_seq = json.loads(date_seq)  # noqa: PLW2901
                # convert date to days since obersavation format
                d = pd.DataFrame().from_dict(date_seq, orient="index")
                d = d[0].apply(
                    lambda x: (
                        datetime(int(str(x)[:4]), int(str(x)[4:6]), int(str(x)[6:]), tzinfo=timezone.utc)
                        - self.reference_date
                    ).days
                )
                date_table.loc[pid, d.values] = 1
            date_table = date_table.fillna(0)
            self.date_tables[s] = {
                index: np.array(list(d.values())) for index, d in date_table.to_dict(orient="index").items()
            }

        # selects patches correspondig to selected folds
        if folds is not None:
            self.meta_patch = pd.concat([self.meta_patch[self.meta_patch["Fold"] == f] for f in folds])

        self.len = self.meta_patch.shape[0]
        self.id_patches = self.meta_patch.index

        # loads normalization values
        if norm:
            self.norm = {}
            for s in self.satellites:
                with open(os.path.join(data_root, f"NORM_{s}_patch.json")) as file:
                    normvals = json.loads(file.read())
                selected_folds = folds if folds is not None else range(1, 6)
                means = [normvals[f"Fold_{f}"]["mean"] for f in selected_folds]
                stds = [normvals[f"Fold_{f}"]["std"] for f in selected_folds]
                self.norm[s] = np.stack(means).mean(axis=0), np.stack(stds).mean(axis=0)
                self.norm[s] = (
                    self.norm[s][0],
                    self.norm[s][1],
                )
        else:
            self.norm = None

    def __len__(self):
        return self.len

    def get_dates(self, id_patch, sat):
        return self.date_range[np.where(self.date_tables[sat][id_patch] == 1)[0]]

    def __getitem__(self, item):
        id_patch = self.id_patches[item]
        output = {}
        satellites = {}
        for satellite in self.satellites:
            data = np.load(
                os.path.join(
                    self.data_root,
                    f"DATA_{satellite}",
                    f"{satellite}_{id_patch}.npy",
                )
            ).astype(np.float32)

            if self.norm is not None:
                data = data - self.norm[satellite][0][None, :, None, None]
                data = data / self.norm[satellite][1][None, :, None, None]

            if self.truncate_image and data.shape[0] > self.truncate_image:
                data = data[-self.truncate_image :]

            if self.pad_image and data.shape[0] < self.pad_image:
                data = pad_numpy(data, self.pad_image)

            satellites[satellite] = data.astype(np.float32)

        if self.target == "semantic":
            target = np.load(os.path.join(self.data_root, "ANNOTATIONS", f"TARGET_{id_patch}.npy"))
            target = target[0].astype(int)
            if self.class_mapping is not None:
                target = self.class_mapping(target)
        elif self.target == "instance":
            heatmap = np.load(os.path.join(self.data_root, "INSTANCE_ANNOTATIONS", f"HEATMAP_{id_patch}.npy"))
            instance_ids = np.load(os.path.join(self.data_root, "INSTANCE_ANNOTATIONS", f"INSTANCES_{id_patch}.npy"))
            zones_path = os.path.join(self.data_root, "INSTANCE_ANNOTATIONS", f"ZONES_{id_patch}.npy")
            pixel_to_object_mapping = np.load(zones_path)
            pixel_semantic_annotation = np.load(os.path.join(self.data_root, "ANNOTATIONS", f"TARGET_{id_patch}.npy"))

            if self.class_mapping is not None:
                pixel_semantic_annotation = self.class_mapping(pixel_semantic_annotation[0])
            else:
                pixel_semantic_annotation = pixel_semantic_annotation[0]

            size = np.zeros((*instance_ids.shape, 2))
            object_semantic_annotation = np.zeros(instance_ids.shape)
            for instance_id in np.unique(instance_ids):
                if instance_id != 0:
                    h = (instance_ids == instance_id).any(axis=-1).sum()
                    w = (instance_ids == instance_id).any(axis=-2).sum()
                    size[pixel_to_object_mapping == instance_id] = (h, w)
                    semantic_value = pixel_semantic_annotation[instance_ids == instance_id][0]
                    object_semantic_annotation[pixel_to_object_mapping == instance_id] = semantic_value

            target = np.concatenate(
                [
                    heatmap[:, :, None],
                    instance_ids[:, :, None],
                    pixel_to_object_mapping[:, :, None],
                    size,
                    object_semantic_annotation[:, :, None],
                    pixel_semantic_annotation[:, :, None],
                ],
                axis=-1,
            ).astype(np.float32)

        dates = {}
        for satellite in self.satellites:
            date = np.array(self.get_dates(id_patch, satellite))

            if self.truncate_image and len(date) > self.truncate_image:
                date = date[-self.truncate_image :]

            if self.pad_image and len(date) < self.pad_image:
                date = pad_dates_numpy(date, self.pad_image)

            dates[satellite] = torch.from_numpy(date)

        output["image"] = satellites["S2"].transpose(0, 2, 3, 1)
        output["mask"] = target

        if self.transform:
            output = self.transform(**output)

        output.update(satellites)
        output["dates"] = dates

        return output

    def plot(self, sample, suptitle=None, show_axes=False):
        dates = sample["dates"]
        target = sample["target"]

        if "S2" not in sample:
            warnings.warn("No RGB image.", stacklevel=2)
            return None

        image_data = sample["S2"]
        date_data = dates["S2"]

        rgb_images = []
        for i in range(image_data.shape[0]):
            rgb_image = image_data[i, :3, :, :].numpy().transpose(1, 2, 0)

            rgb_min = rgb_image.min(axis=(0, 1), keepdims=True)
            rgb_max = rgb_image.max(axis=(0, 1), keepdims=True)
            denom = rgb_max - rgb_min
            denom[denom == 0] = 1
            rgb_image = (rgb_image - rgb_min) / denom

            rgb_images.append(np.clip(rgb_image, 0, 1))

        return self._plot_sample(rgb_images, date_data, target, suptitle=suptitle, show_axes=show_axes)

    def _plot_sample(
        self,
        images: list[np.ndarray],
        dates: torch.Tensor,
        target: torch.Tensor | None,
        suptitle: str | None = None,
        show_axes: bool | None = False,
    ):
        num_images = len(images)
        cols = 5
        rows = (num_images + cols) // cols

        fig, ax = plt.subplots(rows, cols, figsize=(20, 4 * rows))
        axes_visibility = "on" if show_axes else "off"

        for i, image in enumerate(images):
            ax[i // cols, i % cols].imshow(image)
            ax[i // cols, i % cols].set_title(f"Image {i + 1} - Day {dates[i].item()}")
            ax[i // cols, i % cols].axis(axes_visibility)

        if target is not None:
            if rows * cols > num_images:
                target_ax = ax[(num_images) // cols, (num_images) % cols]
            else:
                fig.add_subplot(rows + 1, 1, 1)
                target_ax = fig.gca()

            target_ax.imshow(target.numpy(), cmap="tab20")
            target_ax.set_title("Target")
            target_ax.axis(axes_visibility)

        for k in range(num_images + 1, rows * cols):
            ax[k // cols, k % cols].axis(axes_visibility)

        if suptitle:
            plt.suptitle(suptitle)

        plt.tight_layout()
        return fig



from collections.abc import Callable, Sequence
from typing import Any

import albumentations as A
from kornia.augmentation import AugmentationSequential
from torchgeo.datamodules import NonGeoDataModule

from terratorch.datamodules.generic_multimodal_data_module import wrap_in_compose_is_list
from terratorch.datasets import HelioNetCDFDataset


class ByPassNormalize(Callable):
    def __init__(self):
        super().__init__()

    def __call__(self, batch):
        return batch


class HelioNetCDFDataModule(NonGeoDataModule):
    """NonGeo LightningDataModule implementation for the Heliophysics datamodule."""

    def __init__(
        self,
        train_index_path: str = None,
        val_index_path: str = None,
        test_index_path: str = None,
        predict_index_path: str = None,
        batch_size: int = 4,
        num_workers: int = 0,
        aug: AugmentationSequential = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(HelioNetCDFDataset, batch_size, num_workers, **kwargs)

        self.train_index_path = train_index_path
        self.val_index_path = val_index_path
        self.test_index_path = test_index_path
        self.predict_index_path = predict_index_path

        self.aug = ByPassNormalize() if aug is None else aug
        self.extra_arguments = kwargs

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either fit, validate, test, or predict.
        """
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(
                phase="train",
                index_path=self.train_index_path,
                **self.extra_arguments,
            )

        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(phase="test", index_path=self.val_index_path, **self.extra_arguments)

        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                phase="test", index_path=self.test_index_path, **self.extra_arguments
            )

        if stage in ["predict"]:
            self.predict_dataset = self.dataset_class(
                phase="predict",
                index_path=self.predict_index_path,
                **self.extra_arguments,
            )



import logging
import os
import random
from functools import cache
from logging import Logger

import hdf5plugin  # Enables LZ4 decompression via h5py - does not interfere with gzip
import numpy as np
import pandas as pd
import skimage.measure
import torch
import xarray as xr
import yaml
from numba import njit, prange

try:
    from terratorch_surya.utils.distributed import get_rank
    from terratorch_surya.utils.log import create_logger
except ImportError:
    raise ImportError("It is necessary to install the package `terratorch_surya` to use this module.")
from torch.utils.data import Dataset


@njit(parallel=True)
def fast_transform(data, means, stds, sl_scale_factors, epsilons):
    """
    implements signum log transform using numba for speed
        note: this must reside outside the class definition from which it is called
    """
    C, H, W = data.shape
    out = np.empty((C, H, W), dtype=np.float32)
    for c in prange(C):
        mean = means[c]
        std = stds[c]
        eps = epsilons[c]
        sl_scale_factor = sl_scale_factors[c]
        for i in range(H):
            for j in range(W):
                val = data[c, i, j]
                val = val * sl_scale_factor
                if val >= 0:
                    val = np.log1p(val)
                else:
                    val = -np.log1p(-val)
                out[c, i, j] = (val - mean) / (std + eps)
    return out


class RandomChannelMaskerTransform:
    def __init__(self, num_channels, num_mask_aia_channels, phase, drop_hmi_probablity):
        """
        Initialize the RandomChannelMaskerTransform class as a transform.

        Args:
        - num_channels: Total number of channels in the input (3rd dimension of the tensor).
        - num_mask_aia_channels: Number of channels to randomly mask.
        """
        self.num_channels = num_channels
        self.num_mask_aia_channels = num_mask_aia_channels
        self.drop_hmi_probablity = drop_hmi_probablity

        # print(f'[{phase}] Using Random Channel Masker with num_mask_aia_channels = {self.num_mask_aia_channels}')

    def __call__(self, input_tensor):
        C, T, H, W = input_tensor.shape  # Unpacking the correct 5 dimensions

        # Randomly select channels to mask
        channels_to_mask = random.sample(range(C), self.num_mask_aia_channels)

        # Create an in-place mask of shape [1, 1, num_channels, 1, 1]
        mask = torch.ones((C, 1, 1, 1))
        mask[channels_to_mask, ...] = 0  # Set selected channels to zero

        # Apply the mask in-place for memory efficiency
        masked_tensor = input_tensor * mask  # Modify input_tensor directly

        if self.drop_hmi_probablity > random.random():
            masked_tensor[-1, ...] = 0

        return masked_tensor


class Scaler:
    def __init__(self, var=None):
        self.protected_variables = ["max", "mean", "min", "std", "epsilon", "sl_scale_factor"]

        for k, v in var.items():
            if k in self.protected_variables:
                setattr(self, k, v)


class HelioNetCDFDataset(Dataset):
    """
    PyTorch dataset to load a curated dataset from the NASA Solar Dynamics
    Observatory (SDO) mission stored as NetCDF files, with handling for variable timesteps.

    Internally maintains two databases. The first is`self.index`. This takes the form
                                                                        path  present
        timestep
        2011-01-01 00:00:00  /lustre/fs0/scratch/shared/data/2011/01/Arka_2...        1
        2011-01-01 00:12:00  /lustre/fs0/scratch/shared/data/2011/01/Arka_2...        1
        ...                                                                ...      ...
        2012-11-30 23:48:00  /lustre/fs0/scratch/shared/data/2012/11/Arka_2...        1

    The second is `self.valid_indices`. This is simply a list of timesteps -- entries in the
    index of `self.index` -- which define valid samples. A sample is valid when all timestamps
    that can be reached by entries in time_delta_input_minutes and time_delta_target_minutes
    can be reached from it are present.
    """

    def __init__(
        self,
        index_path: str,
        time_delta_input_minutes: list[int],
        time_delta_target_minutes: int,
        n_input_timestamps: int,
        rollout_steps: int,
        scalers=None,
        num_mask_aia_channels=0,
        drop_hmi_probablity=0,
        use_latitude_in_learned_flow=False,
        channels: list[str] | None = None,
        phase="train",
        pooling: int | None = None,
        initial_step: str = None,
        random_vert_flip: bool = False,
    ):
        # If the argument `scalers` is a path to a json file
        # it must be adapted to work as a dictionary of Scaler objects
        if isinstance(scalers, str):
            with open(scalers) as fp:
                scalers = yaml.safe_load(fp)

            scalers_classes = {}
            for k, v in scalers.items():
                scalers_classes[k] = Scaler(var=v)

            self.scalers = scalers_classes

        elif isinstance(scalers, dict):
            scalers_classes = {}
            for k, v in scalers.items():
                scalers_classes[k] = Scaler(var=v)

            self.scalers = scalers_classes

        else:
            raise ValueError(f"`scalers` is expected to be `str` or `dict`, but received {type(self.scalers)}")

        self.phase = phase
        self.channels = channels
        self.num_mask_aia_channels = num_mask_aia_channels
        self.drop_hmi_probablity = drop_hmi_probablity
        self.n_input_timestamps = n_input_timestamps
        self.rollout_steps = rollout_steps
        self.use_latitude_in_learned_flow = use_latitude_in_learned_flow
        self.pooling = pooling if pooling is not None else 1
        self.random_vert_flip = random_vert_flip

        if self.channels is None:
            # AIA + HMI channels
            self.channels = [
                "0094",
                "0131",
                "0171",
                "0193",
                "0211",
                "0304",
                "0335",
                "hmi",
            ]
        self.in_channels = len(self.channels)

        self.masker = RandomChannelMaskerTransform(
            num_channels=self.in_channels,
            num_mask_aia_channels=self.num_mask_aia_channels,
            phase=self.phase,
            drop_hmi_probablity=self.drop_hmi_probablity,
        )

        # Convert time delta to numpy timedelta64
        self.time_delta_input_minutes = sorted(np.timedelta64(t, "m") for t in time_delta_input_minutes)

        self.time_delta_target_minutes = [
            np.timedelta64(iroll * time_delta_target_minutes, "m") for iroll in range(1, rollout_steps + 2)
        ]

        # Create the index
        self.root_path = os.path.dirname(index_path)
        self.index = pd.read_csv(index_path)
        self.index = self.index[self.index["present"] == 1]
        self.index["timestep"] = pd.to_datetime(self.index["timestep"]).values.astype("datetime64[ns]")
        self.index.set_index("timestep", inplace=True)
        self.index.sort_index(inplace=True)

        # Filter out rows where the sequence is not fully present
        self.valid_indices = self.filter_valid_indices()
        self.adjusted_length = len(self.valid_indices)

        self.rank = get_rank()
        self.logger: Logger | None = None

    def create_logger(self):
        """
        Creates a logger attached to self.logger.
        The logger is identified by SLURM job ID
        as well as the data processes rank and process ID.
        """
        slurm_job_id = os.getenv("SLURM_JOB_ID")
        pid = os.getpid()
        self.logger = create_logger(
            output_dir="logs/data",
            dist_rank=self.rank,
            name=f"{slurm_job_id}_{self.rank:>03}_data_{self.phase}_{pid}",
        )

    def filter_valid_indices(self):
        """
        Extracts timestamps from the index of self.index that define valid
        samples.

        Args:
        Returns:
            List of timestamps.
        """

        valid_indices = []
        time_deltas = np.unique(self.time_delta_input_minutes + self.time_delta_target_minutes)

        for reference_timestep in self.index.index:
            required_timesteps = reference_timestep + time_deltas

            if all(t in self.index.index for t in required_timesteps):
                valid_indices.append(reference_timestep)

        return valid_indices

    def __len__(self):
        return self.adjusted_length

    def __getitem__(self, idx: int) -> dict:
        """
        Args:
            idx: Index of sample to load. (Pytorch standard.)
        Returns:
            Dictionary with following keys. The values are tensors with shape as follows:
                ts (torch.Tensor):                C, T, H, W
                time_delta_input (torch.Tensor):  T
                input_latitude (torch.Tensor):    T
                forecast (torch.Tensor):          C, L, H, W
                lead_time_delta (torch.Tensor):   L
                forecast_latitude (torch.Tensor): L
            C - Channels, T - Input times, H - Image height, W - Image width, L - Lead time.
        """

        if self.logger is None:
            try:
                self.create_logger()
            except FileNotFoundError:
                self.logger = logging.getLogger("terratorch")

        exception_counter = 0
        max_exception = 100

        self.logger.info(f"Starting to retrieve index {idx}.")

        sample = self._get_index_data(idx)

        while True:
            try:
                sample = self._get_index_data(idx)
            except Exception as e:
                exception_counter += 1
                if exception_counter >= max_exception:
                    raise e

                reference_timestep = self.valid_indices[idx]
                self.logger.warning(
                    f"Failed retrieving index {idx}. Timestamp {reference_timestep}. Attempt {exception_counter}."
                )

                idx = (idx + 1) % self.__len__()
            else:
                self.logger.info(f"Returning index {idx}.")
                return sample

    def _get_index_data(self, idx: int) -> dict:
        """
        Args:
            idx: Index of sample to load. (Pytorch standard.)
        Returns:
            Dictionary with following keys. The values are tensors with shape as follows:
                ts (torch.Tensor):                C, T, H, W
                time_delta_input (torch.Tensor):  T
                input_latitude (torch.Tensor):    T
                forecast (torch.Tensor):          C, L, H, W
                lead_time_delta (torch.Tensor):   L
                forecast_latitude (torch.Tensor): L
            C - Channels, T - Input times, H - Image height, W - Image width, L - Lead time.
        """
        # start_time = time.time()

        time_deltas = np.array(
            sorted(random.sample(self.time_delta_input_minutes[:-1], self.n_input_timestamps - 1))
            + [self.time_delta_input_minutes[-1]]
            + self.time_delta_target_minutes
        )

        reference_timestep = self.valid_indices[idx]
        required_timesteps = reference_timestep + time_deltas

        sequence_data = [
            self.transform_data(self.load_nc_data(self.index.loc[timestep, "path"], timestep, self.channels))
            for timestep in required_timesteps
        ]

        # Split sequence_data into inputs and target
        inputs = sequence_data[: -self.rollout_steps - 1]
        targets = sequence_data[-self.rollout_steps - 1 :]

        stacked_inputs = np.stack(inputs, axis=1)
        stacked_targets = np.stack(targets, axis=1)

        timestamps_input = required_timesteps[: -self.rollout_steps - 1]
        timestamps_targets = required_timesteps[-self.rollout_steps - 1 :]

        if self.num_mask_aia_channels > 0 or self.drop_hmi_probablity:
            # assert 0 < self.num_mask_aia_channels < self.in_channels, \
            #     f'num_mask_aia_channels = {self.num_mask_aia_channels} should lie between 0 and {self.in_channels}'

            stacked_inputs = self.masker(stacked_inputs)

        time_delta_input_float = (
            time_deltas[-self.rollout_steps - 2] - time_deltas[: -self.rollout_steps - 1]
        ) / np.timedelta64(1, "h")
        time_delta_input_float = time_delta_input_float.astype(np.float32)

        lead_time_delta_float = (
            time_deltas[-self.rollout_steps - 2] - time_deltas[-self.rollout_steps - 1 :]
        ) / np.timedelta64(1, "h")
        lead_time_delta_float = lead_time_delta_float.astype(np.float32)

        # print('LocalRank', int(os.environ["LOCAL_RANK"]),
        #       'GlobalRank', int(os.environ["RANK"]),
        #       'worker', torch.utils.data.get_worker_info().id,
        #       f': Processed Input: {idx} ',time.time()- start_time)

        metadata = {
            "timestamps_input": timestamps_input,
            "timestamps_targets": timestamps_targets,
        }

        if self.random_vert_flip:
            if torch.bernoulli(torch.ones(()) / 2) == 1:
                stacked_inputs = torch.flip(stacked_inputs, dims=-2)
                stacked_targets = torch.flip(stacked_inputs, dims=-2)

        if self.use_latitude_in_learned_flow:
            from sunpy.coordinates.ephemeris import get_earth

            sequence_latitude = [get_earth(timestep).lat.value for timestep in required_timesteps]
            input_latitudes = sequence_latitude[: -self.rollout_steps - 1]
            target_latitude = sequence_latitude[-self.rollout_steps - 1 :]

            return {
                "ts": stacked_inputs,
                "time_delta_input": time_delta_input_float,
                "input_latitudes": input_latitudes,
                "forecast": stacked_targets,
                "lead_time_delta": lead_time_delta_float,
                "forecast_latitude": target_latitude,
            }, metadata

        if self.phase == "predict":
            return {
                "ts": stacked_inputs,
                "time_delta_input": time_delta_input_float,
                "forecast": stacked_targets,
                "lead_time_delta": lead_time_delta_float,
            }

        else:
            return {
                "ts": stacked_inputs,
                "time_delta_input": time_delta_input_float,
                "forecast": stacked_targets,
                "lead_time_delta": lead_time_delta_float,
            }, metadata

    def load_nc_data(self, filepath: str, timestep: pd.Timestamp, channels: list[str]) -> np.ndarray:
        """
        Args:
            filepath: String or Pathlike. Points to NetCDF file to open.
            timestep: Identifies timestamp to retrieve.
        Returns:
            Numpy array of shape (C, H, W).
        """

        self.logger.info(f"Reading file {filepath}.")
        with xr.open_dataset(filepath, engine="h5netcdf", chunks=None, cache=False) as ds:
            data = ds[channels].to_array().values
        return data

    @cache
    def transformation_inputs(self) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray):
        means = np.array([self.scalers[ch].mean for ch in self.channels])
        stds = np.array([self.scalers[ch].std for ch in self.channels])
        epsilons = np.array([self.scalers[ch].epsilon for ch in self.channels])
        sl_scale_factors = np.array([self.scalers[ch].sl_scale_factor for ch in self.channels])

        return means, stds, epsilons, sl_scale_factors

    def transform_data(self, data: np.ndarray) -> np.ndarray:
        """
        Applies scalers.

        Args:
            data: Numpy array of shape (C, H, W)
        Returns:
            Tensor of shape (C, H, W). Data type float32.
        Uses:
                numba to speed up transform
                tvk-srm-heliofm  environment cloned from srm-heliofm with numba added
                tvk_dgx_slurm.sh  shell script modified to use new environment and new jobname
                train_spectformer_dgx.yaml new jobname
        """
        assert data.ndim == 3

        if self.pooling > 1:
            data = skimage.measure.block_reduce(data, block_size=(1, self.pooling, self.pooling), func=np.mean)

        means, stds, epsilons, sl_scale_factors = self.transformation_inputs()
        result_np = fast_transform(data, means, stds, sl_scale_factors, epsilons)
        return result_np



import torch
from torch import nn

import os
import typing 
import logging
import importlib

from terratorch.models.model import (
    Model,
    ModelFactory,
    ModelOutput,
)
from terratorch.registry import MODEL_FACTORY_REGISTRY

from terratorch.models.pincers.unet_pincer import UNetPincer
from terratorch.models.pincers.wxc_downscaling_pincer import get_downscaling_pincer


logger = logging.getLogger(__name__)

class WxCModuleWrapper(Model, nn.Module):
    def __init__(self, module: nn.Module) -> None:
        super().__init__()
        self.module = module
        
    def freeze_encoder(self):
        logger.info("freeze encoder")
        for param in self.module.backbone.parameters():
            param.requires_grad = False

    def freeze_decoder(self):
        logger.info("freeze decoder")
        for param in self.module.head.parameters():
            param.requires_grad = False

    def forward(self, x) -> ModelOutput:
        mo = self.module.forward(x)
        return ModelOutput(mo)
    
    def load_state_dict(self, state_dict: os.Mapping[str, typing.Any], strict: bool = True, assign: bool = False):
        self.module.load_state_dict(state_dict, strict, assign)

@MODEL_FACTORY_REGISTRY.register
class WxCModelFactory(ModelFactory):
    def build_model(
        self,
        backbone: str | nn.Module,
        aux_decoders: str,
        checkpoint_path:str=None,
        backbone_weights: str = None,
        **kwargs,
    ) -> Model:
        if backbone == 'prithviwxc':
            try:
                prithviwxc = importlib.import_module('PrithviWxC.model')
            except ModuleNotFoundError as e:
                print(f"Module not found: {e.name}. Please install PrithviWxC using pip install PrithviWxC")
                raise

            #remove parameters not meant for the backbone but for other parts of the model
            logger.debug(kwargs)
            if 'skip_connection' in kwargs.keys():
                skip_connection = kwargs.pop('skip_connection')
            else:
                skip_connection = True
            if 'wxc_auxiliary_data_path' in kwargs.keys():
                wxc_auxiliary_data_path = kwargs.pop('wxc_auxiliary_data_path')
            if 'config_path' in kwargs.keys():
                config_path = kwargs.pop('config_path')
                # from granitewxc.utils.config import get_config #TODO rkie fix: import flaky
                from granitewxc.utils.config import ExperimentConfig
                import yaml
                def get_config(config_path: str) -> ExperimentConfig:
                    cfg = yaml.safe_load(open(config_path, 'r'))
                    return ExperimentConfig.from_dict(cfg)

                # end TODO
                config = get_config(config_path)

            if aux_decoders == 'unetpincer':
                backbone = prithviwxc.PrithviWxC(**kwargs)
                # Freeze PrithviWxC model parameters
                for param in backbone.parameters():
                    param.requires_grad = False

                # Load pre-trained weights if checkpoint is provided
                if backbone_weights is not None:

                    print(f"Starting to load model from {backbone_weights}")
                    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
                    state_dict = torch.load(
                        f=backbone_weights,
                        weights_only=True,
                        map_location=torch.device(device),
                    )

                    # Compare the keys in model and saved state_dict
                    model_keys = set(backbone.state_dict().keys())
                    saved_state_dict_keys = set(state_dict.keys())

                    # Find keys that are in the model but not in the saved state_dict
                    missing_in_saved = model_keys - saved_state_dict_keys
                    # Find keys that are in the saved state_dict but not in the model
                    missing_in_model = saved_state_dict_keys - model_keys
                    # Find keys that are common between the model and the saved state_dict
                    common_keys = model_keys & saved_state_dict_keys

                    # Print the common keys
                    if common_keys:
                        print(f"Keys loaded : {common_keys}")

                    # Print the discrepancies
                    if missing_in_saved:
                        print(f"Keys present in model but missing in saved state_dict: {missing_in_saved}")
                    if missing_in_model:
                        print(f"Keys present in saved state_dict but missing in model: {missing_in_model}")

                    # Load the state_dict with strict=False to allow partial loading
                    backbone.load_state_dict(state_dict=state_dict, strict=False)
                    print('=>' * 10, f"Model loaded from {backbone_weights}...")
                    print("Loaded backbone weights")
                else:
                    print('Not loading backbone model weigts')

                device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
                backbone.to(device)
                model_to_return = UNetPincer(backbone, skip_connection=skip_connection).to(device)
                return model_to_return
            if aux_decoders == 'downscaler':
                try:
                    __import__('granitewxc.utils.config')
                    from granitewxc.utils.config import get_config
                    from granitewxc.utils.downscaling_model import get_backbone, get_finetune_model
                    config.data.data_path_surface = os.path.join(wxc_auxiliary_data_path, 'merra-2')
                    config.data.data_path_vertical = os.path.join(wxc_auxiliary_data_path, 'merra-2')
                    config.data.climatology_path_surface = os.path.join(wxc_auxiliary_data_path, 'climatology')
                    config.data.climatology_path_vertical = os.path.join(wxc_auxiliary_data_path, 'climatology')
                    config.model.input_scalers_surface_path = os.path.join(wxc_auxiliary_data_path, 'climatology/musigma_surface.nc')
                    config.model.input_scalers_vertical_path = os.path.join(wxc_auxiliary_data_path, 'climatology/musigma_vertical.nc')
                    config.model.output_scalers_surface_path = os.path.join(wxc_auxiliary_data_path, 'climatology/anomaly_variance_surface.nc')
                    config.model.output_scalers_vertical_path = os.path.join(wxc_auxiliary_data_path, 'climatology/anomaly_variance_vertical.nc')
                    backbone = get_backbone(config)
                except ImportError as e:
                    print(f"Module not found: {e.name}. Please install granitewxc using pip install granitewxc")
                dsp = get_downscaling_pincer(config, backbone)
                if checkpoint_path:
                    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
                    dsp.load_state_dict(torch.load(checkpoint_path, weights_only=False, map_location=device))
                return dsp
            return WxCModuleWrapper(backbone)

        elif backbone == 'prithvi-eccc-downscaling':
            try:
                model_args = kwargs['model_args']
                if model_args.model.unet:
                    from granitewxc.models.model import get_finetune_model_UNET
                    module = get_finetune_model_UNET(model_args)
                else:
                    from granitewxc.models.model import get_finetune_model
                    module = get_finetune_model(model_args)

                if checkpoint_path:
                    weights_path = model_args.path_model_weights
                    weights = torch.load(weights_path)['model']
                    try:
                        module.load_state_dict(weights)
                    except Exception as e:
                        # remove the module prefix from the weights (for models trained with DDP)
                        weights = {key.replace('model.module.', ''): value for key, value in weights.items()}
                        module.load_state_dict(weights)

                if model_args.path_backbone_weights:
                    module.load_pretrained_backbone(
                        weights_path = model_args.path_backbone_weights,
                        sel_prefix=model_args.backbone_prefix,
                        freeze = model_args.backbone_freeze,
                    )

                return WxCModuleWrapper(module)
            except ImportError:
                print("ECCC downscaling module not found")
                print("To install it, run:")
                print("pip install git+https://github.com/IBM/granite-wxc.git")

        # starting from there only for backwards compatibility, deprecated
        if backbone == 'gravitywave':
            try:
                __import__('prithviwxc.gravitywave.inference')
                from prithviwxc.gravitywave.inference import get_model
                from prithviwxc.gravitywave.config import get_cfg
                cfg = get_cfg()
                model_wrapper = WxCModuleWrapper(get_model(cfg,'uvtp122', cfg.singular_sharded_checkpoint))
                if checkpoint_path:
                    model_wrapper.load_state_dict(torch.load(checkpoint_path, weights_only=True))
                return model_wrapper
            except ImportError as e:
                missing_module = e.name if hasattr(e, 'name') else "unknown module"
                print('prithvi wxc gravitywave not installed, missing module: {missing_module}')
                return None
        else:
            try:
                __import__('granitewxc.utils.config')
                from granitewxc.utils.config import get_config
                from granitewxc.utils.downscaling_model import get_finetune_model
                module = get_finetune_model(kwargs['model_config'])
                model_wrapper = WxCModuleWrapper(module)
                
                if checkpoint_path:
                    model_wrapper.load_state_dict(torch.load(checkpoint_path, weights_only=True))
                return model_wrapper
            except ImportError:
                print('granite wxc downscaling not installed')



import os
from typing import Optional

import torch
from torch.utils.data import DataLoader
from torchvision.datasets import CocoDetection
import torchvision.transforms as T
import lightning as pl




class ElephantCocoDataset(CocoDetection):
    def __init__(self, img_folder, ann_file, transform=None):
        super().__init__(img_folder, ann_file)
        self.transform = transform

    def __getitem__(self, idx):
        img, target = super().__getitem__(idx)


        # convert COCO target to detection-style dict
        boxes = []
        labels = []

        """ print('_+_+_+_+_+')
        print('getimtem') """

        for obj in target:
            x, y, w, h = obj["bbox"]

            # strict filter: positive size only
            if w > 0 and h > 0:
                x2, y2 = x + w, y + h
                if x2 > x and y2 > y:  # extra guard
                    """ print('_+_+_+_+_+')
                    print("appending box")
                    print([x, y, x2, y2]) """
                    boxes.append([x, y, x2, y2])
                    labels.append(obj["category_id"])



        if len (boxes) > 0:
            boxes = torch.tensor(boxes, dtype=torch.float32)
            labels = torch.tensor(labels, dtype=torch.int64)
        else:
            boxes = torch.zeros((0, 4), dtype=torch.float32)
            labels = torch.zeros((0,), dtype=torch.int64)


        """ if len(boxes) == 0:
            # IMPORTANT: models like FasterRCNN break if no boxes are returned
            # safest is to skip this sample entirely
            return self.__getitem__((idx + 1) % len(self)) """
        
        if self.transform:
            img = self.transform(img)

        to_tensor = T.ToTensor()

        sample = {
            "boxes": boxes,
            "labels": labels,
            "image": to_tensor(img),
            "image_id": torch.tensor([self.ids[idx]]),
        }

        return sample